
# FIXME this is named wrong!

import vstruct
from vstruct.primitives import *

class CLIENT_ID(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.UniqueProcess = v_ptr()
        self.UniqueThread = v_ptr()

class EXCEPTION_RECORD(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.ExceptionCode = v_uint32()
        self.ExceptionFlags = v_uint32()
        self.ExceptionRecord = v_ptr()
        self.ExceptionAddress = v_ptr()
        self.NumberParameters = v_uint32()

class EXCEPTION_REGISTRATION(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.prev = v_ptr()
        self.handler = v_ptr()

class HEAP(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.Entry = HEAP_ENTRY()
        self.Signature = v_uint32()
        self.Flags = v_uint32()
        self.ForceFlags = v_uint32()
        self.VirtualMemoryThreshold = v_uint32()
        self.SegmentReserve = v_uint32()
        self.SegmentCommit = v_uint32()
        self.DeCommitFreeBlockThreshold = v_uint32()
        self.DeCommitTotalFreeThreshold = v_uint32()
        self.TotalFreeSize = v_uint32()
        self.MaximumAllocationSize = v_uint32()
        self.ProcessHeapsListIndex = v_uint16()
        self.HeaderValidateLength = v_uint16()
        self.HeaderValidateCopy = v_ptr()
        self.NextAvailableTagIndex = v_uint16()
        self.MaximumTagIndex = v_uint16()
        self.TagEntries = v_ptr()
        self.UCRSegments = v_ptr()
        self.UnusedUnCommittedRanges = v_ptr()
        self.AlignRound = v_uint32()
        self.AlignMask = v_uint32()
        self.VirtualAllocBlocks = ListEntry()
        self.Segments = vstruct.VArray([v_uint32() for i in range(64)])
        self.u = vstruct.VArray([v_uint8() for i in range(16)])
        self.u2 = vstruct.VArray([v_uint8() for i in range(2)])
        self.AllocatorBackTraceIndex = v_uint16()
        self.NonDedicatedListLength = v_uint32()
        self.LargeBlocksIndex = v_ptr()
        self.PseudoTagEntries = v_ptr()
        self.FreeLists = vstruct.VArray([ListEntry() for i in range(128)])
        self.LockVariable = v_uint32()
        self.CommitRoutine = v_ptr()
        self.FrontEndHeap = v_ptr()
        self.FrontEndHeapLockCount = v_uint16()
        self.FrontEndHeapType = v_uint8()
        self.LastSegmentIndex = v_uint8()

class HEAP_SEGMENT(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.Entry = HEAP_ENTRY()
        self.SegmentSignature = v_uint32()
        self.SegmentFlags = v_uint32()
        self.Heap = v_ptr()
        self.LargestUncommitedRange = v_uint32()
        self.BaseAddress = v_ptr()
        self.NumberOfPages = v_uint32()
        self.FirstEntry = v_ptr()
        self.LastValidEntry = v_ptr()
        self.NumberOfUnCommittedPages = v_uint32()
        self.NumberOfUnCommittedRanges = v_uint32()
        self.UncommittedRanges = v_ptr()
        self.SegmentAllocatorBackTraceIndex = v_uint16()
        self.Reserved = v_uint16()
        self.LastEntryInSegment = v_ptr()

class HEAP_ENTRY(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.Size = v_uint16()
        self.PrevSize = v_uint16()
        self.SegmentIndex = v_uint8()
        self.Flags = v_uint8()
        self.Unused = v_uint8()
        self.TagIndex = v_uint8()

class ListEntry(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.Flink = v_ptr()
        self.Blink = v_ptr()

class NT_TIB(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.ExceptionList = v_ptr()
        self.StackBase = v_ptr()
        self.StackLimit = v_ptr()
        self.SubSystemTib = v_ptr()
        self.FiberData = v_ptr()
        #x.Version = v_ptr() # This is a union field
        self.ArbitraryUserPtr = v_ptr()
        self.Self = v_ptr()

class PEB(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.InheritedAddressSpace = v_uint8()
        self.ReadImageFileExecOptions = v_uint8()
        self.BeingDebugged = v_uint8()
        self.SpareBool = v_uint8()
        self.Mutant = v_ptr()
        self.ImageBaseAddress = v_ptr()
        self.Ldr = v_ptr()
        self.ProcessParameters = v_ptr()
        self.SubSystemData = v_ptr()
        self.ProcessHeap = v_ptr()
        self.FastPebLock = v_ptr()
        self.FastPebLockRoutine = v_ptr()
        self.FastPebUnlockRoutine = v_ptr()
        self.EnvironmentUpdateCount = v_uint32()
        self.KernelCallbackTable = v_ptr()
        self.SystemReserved = v_uint32()
        self.AtlThunkSListPtr32 = v_ptr()
        self.FreeList = v_ptr()
        self.TlsExpansionCounter = v_uint32()
        self.TlsBitmap = v_ptr()
        self.TlsBitmapBits = vstruct.VArray([v_uint32() for i in range(2)])
        self.ReadOnlySharedMemoryBase = v_ptr()
        self.ReadOnlySharedMemoryHeap = v_ptr()
        self.ReadOnlyStaticServerData = v_ptr()
        self.AnsiCodePageData = v_ptr()
        self.OemCodePageData = v_ptr()
        self.UnicodeCaseTableData = v_ptr()
        self.NumberOfProcessors = v_uint32()
        self.NtGlobalFlag = v_uint64()
        self.CriticalSectionTimeout = v_uint64()
        self.HeapSegmentReserve = v_uint32()
        self.HeapSegmentCommit = v_uint32()
        self.HeapDeCommitTotalFreeThreshold = v_uint32()
        self.HeapDeCommitFreeBlockThreshold = v_uint32()
        self.NumberOfHeaps = v_uint32()
        self.MaximumNumberOfHeaps = v_uint32()
        self.ProcessHeaps = v_ptr()
        self.GdiSharedHandleTable = v_ptr()
        self.ProcessStarterHelper = v_ptr()
        self.GdiDCAttributeList = v_uint32()
        self.LoaderLock = v_ptr()
        self.OSMajorVersion = v_uint32()
        self.OSMinorVersion = v_uint32()
        self.OSBuildNumber = v_uint16()
        self.OSCSDVersion = v_uint16()
        self.OSPlatformId = v_uint32()
        self.ImageSubsystem = v_uint32()
        self.ImageSubsystemMajorVersion = v_uint32()
        self.ImageSubsystemMinorVersion = v_uint32()
        self.ImageProcessAffinityMask = v_uint32()
        self.GdiHandleBuffer = vstruct.VArray([v_ptr() for i in range(34)])
        self.PostProcessInitRoutine = v_ptr()
        self.TlsExpansionBitmap = v_ptr()
        self.TlsExpansionBitmapBits = vstruct.VArray([v_uint32() for i in range(32)])
        self.SessionId = v_uint32()
        self.AppCompatFlags = v_uint64()
        self.AppCompatFlagsUser = v_uint64()
        self.pShimData = v_ptr()
        self.AppCompatInfo = v_ptr()
        self.CSDVersion = v_ptr()
        self.UNKNOWN = v_uint32()
        self.ActivationContextData = v_ptr()
        self.ProcessAssemblyStorageMap = v_ptr()
        self.SystemDefaultActivationContextData = v_ptr()
        self.SystemAssemblyStorageMap = v_ptr()
        self.MinimumStackCommit = v_uint32()

class SEH3_SCOPETABLE(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.EnclosingLevel = v_int32()
        self.FilterFunction = v_ptr()
        self.HandlerFunction = v_ptr()

class SEH4_SCOPETABLE(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.GSCookieOffset = v_int32()
        self.GSCookieXOROffset = v_int32()
        self.EHCookieOffset = v_int32()
        self.EHCookieXOROffset = v_int32()
        self.EnclosingLevel = v_int32()
        self.FilterFunction = v_ptr()
        self.HandlerFunction = v_ptr()

class TEB(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.TIB = NT_TIB()
        self.EnvironmentPointer = v_ptr()
        self.ClientId = CLIENT_ID()
        self.ActiveRpcHandle = v_ptr()
        self.ThreadLocalStorage = v_ptr()
        self.ProcessEnvironmentBlock = v_ptr()
        self.LastErrorValue = v_uint32()
        self.CountOfOwnedCriticalSections = v_uint32()
        self.CsrClientThread = v_ptr()
        self.Win32ThreadInfo = v_ptr()
        self.User32Reserved = vstruct.VArray([v_uint32() for i in range(26)])
        self.UserReserved = vstruct.VArray([v_uint32() for i in range(5)])
        self.WOW32Reserved = v_ptr()
        self.CurrentLocale = v_uint32()
        self.FpSoftwareStatusRegister = v_uint32()

class CLSID(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.uuid = GUID()

class IID(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.uuid = GUID()




from .rflib_defs import *
from .chipcondefs import *
from .rflib_version import *

#### supporting chipcon_usb.py
DEFAULT_USB_TIMEOUT = 1000

EP_TIMEOUT_IDLE     = 400
EP_TIMEOUT_ACTIVE   = 10

USB_MAX_BLOCK_SIZE  = 512
USB_RX_WAIT         = 1000
USB_TX_WAIT         = 10000

USB_BM_REQTYPE_TGTMASK          =0x1f
USB_BM_REQTYPE_TGT_DEV          =0x00
USB_BM_REQTYPE_TGT_INTF         =0x01
USB_BM_REQTYPE_TGT_EP           =0x02

USB_BM_REQTYPE_TYPEMASK         =0x60
USB_BM_REQTYPE_TYPE_STD         =0x00
USB_BM_REQTYPE_TYPE_CLASS       =0x20
USB_BM_REQTYPE_TYPE_VENDOR      =0x40
USB_BM_REQTYPE_TYPE_RESERVED    =0x60

USB_BM_REQTYPE_DIRMASK          =0x80
USB_BM_REQTYPE_DIR_OUT          =0x00
USB_BM_REQTYPE_DIR_IN           =0x80

USB_GET_STATUS                  =0x00
USB_CLEAR_FEATURE               =0x01
USB_SET_FEATURE                 =0x03
USB_SET_ADDRESS                 =0x05
USB_GET_DESCRIPTOR              =0x06
USB_SET_DESCRIPTOR              =0x07
USB_GET_CONFIGURATION           =0x08
USB_SET_CONFIGURATION           =0x09
USB_GET_INTERFACE               =0x0a
USB_SET_INTERFACE               =0x11
USB_SYNCH_FRAME                 =0x12

APP_GENERIC                     = 0x01
APP_DEBUG                       = 0xfe
APP_SYSTEM                      = 0xff


SYS_CMD_PEEK                    = 0x80
SYS_CMD_POKE                    = 0x81
SYS_CMD_PING                    = 0x82
SYS_CMD_STATUS                  = 0x83
SYS_CMD_POKE_REG                = 0x84
SYS_CMD_GET_CLOCK               = 0x85
SYS_CMD_BUILDTYPE               = 0x86
SYS_CMD_BOOTLOADER              = 0x87
SYS_CMD_RFMODE                  = 0x88
SYS_CMD_COMPILER                = 0x89
SYS_CMD_PARTNUM                 = 0x8e
SYS_CMD_RESET                   = 0x8f
SYS_CMD_CLEAR_CODES             = 0x90
SYS_CMD_DEVICE_SERIAL_NUMBER    = 0x91
SYS_CMD_LED_MODE                = 0x93

EP0_CMD_GET_DEBUG_CODES         = 0x00
EP0_CMD_GET_ADDRESS             = 0x01
EP0_CMD_POKEX                   = 0x01
EP0_CMD_PEEKX                   = 0x02
EP0_CMD_PING0                   = 0x03
EP0_CMD_PING1                   = 0x04
EP0_CMD_RESET                   = 0xfe


DEBUG_CMD_STRING                = 0xf0
DEBUG_CMD_HEX                   = 0xf1
DEBUG_CMD_HEX16                 = 0xf2
DEBUG_CMD_HEX32                 = 0xf3
DEBUG_CMD_INT                   = 0xf4

EP5OUT_MAX_PACKET_SIZE          = 64
EP5IN_MAX_PACKET_SIZE           = 64
# EP5OUT_BUFFER_SIZE must match firmware/include/chipcon_usb.h definition
EP5OUT_BUFFER_SIZE              = 516

LC_USB_INITUSB                = 0x2
LC_MAIN_RFIF                  = 0xd
LC_USB_DATA_RESET_RESUME      = 0xa
LC_USB_RESET                  = 0xb
LC_USB_EP5OUT                 = 0xc
LC_RF_VECTOR                  = 0x10
LC_RFTXRX_VECTOR              = 0x11

LCE_USB_EP5_TX_WHILE_INBUF_WRITTEN    = 0x1
LCE_USB_EP0_SENT_STALL                = 0x4
LCE_USB_EP5_OUT_WHILE_OUTBUF_WRITTEN  = 0x5
LCE_USB_EP5_LEN_TOO_BIG               = 0x6
LCE_USB_EP5_GOT_CRAP                  = 0x7
LCE_USB_EP5_STALL                     = 0x8
LCE_USB_DATA_LEFTOVER_FLAGS           = 0x9
LCE_RF_RXOVF                          = 0x10
LCE_RF_TXUNF                          = 0x11

RCS = {}
LCS = {}
LCES = {}
lcls = locals()
for lcl in list(lcls.keys()):
    if lcl.startswith("LCE_"):
        LCES[lcl] = lcls[lcl]
        LCES[lcls[lcl]] = lcl
    if lcl.startswith("LC_"):
        LCS[lcl] = lcls[lcl]
        LCS[lcls[lcl]] = lcl
    if lcl.startswith("RC_"):
        RCS[lcl] = lcls[lcl]
        RCS[lcls[lcl]] = lcl

CHIPS = {
    0x91: "CC2511",
    0x81: "CC2510",
    0x11: "CC1111",
    0x01: "CC1110",
    0x40: "FakeDongle",
    }


#### supporting chipcon_nic.py
# band limits in Hz
FREQ_MIN_300  = 281000000
FREQ_MAX_300  = 361000000
FREQ_MIN_400  = 378000000
FREQ_MAX_400  = 481000000
FREQ_MIN_900  = 749000000
FREQ_MAX_900  = 962000000

# band transition points in Hz
FREQ_EDGE_400 = 369000000
FREQ_EDGE_900 = 615000000

# VCO transition points in Hz
FREQ_MID_300  = 318000000
FREQ_MID_400  = 424000000
FREQ_MID_900  = 848000000

SYNCM_NONE                      = 0
SYNCM_15_of_16                  = 1
SYNCM_16_of_16                  = 2
SYNCM_30_of_32                  = 3
SYNCM_CARRIER                   = 4
SYNCM_CARRIER_15_of_16          = 5
SYNCM_CARRIER_16_of_16          = 6
SYNCM_CARRIER_30_of_32          = 7

RF_SUCCESS                      = 0

RF_MAX_TX_BLOCK                 = 255
RF_MAX_TX_CHUNK                 = 240 # must match MAX_TX_MSGLEN in firmware/include/FHSS.h
                                      # and be divisible by 16 for crypto operations
RF_MAX_TX_LONG                  = 65535
RF_MAX_RX_BLOCK                 = 512 # must match BUFFER_SIZE definition in firmware/include/cc1111rf.h

APP_NIC =                       0x42
APP_SPECAN =                    0x43

NIC_RECV =                      0x1
NIC_XMIT =                      0x2
NIC_SET_ID =                    0x3
NIC_SET_RECV_LARGE =            0x5
NIC_SET_AES_MODE =              0x6
NIC_GET_AES_MODE =              0x7
NIC_SET_AES_IV =                0x8
NIC_SET_AES_KEY =               0x9
NIC_SET_AMP_MODE =              0xa
NIC_GET_AMP_MODE =              0xb
NIC_LONG_XMIT =                 0xc
NIC_LONG_XMIT_MORE =            0xd

FHSS_SET_CHANNELS =             0x10
FHSS_NEXT_CHANNEL =             0x11
FHSS_CHANGE_CHANNEL =           0x12
FHSS_SET_MAC_THRESHOLD =        0x13
FHSS_GET_MAC_THRESHOLD =        0x14
FHSS_SET_MAC_DATA =             0x15
FHSS_GET_MAC_DATA =             0x16
FHSS_XMIT =                     0x17
FHSS_GET_CHANNELS =             0x18

FHSS_SET_STATE =                0x20
FHSS_GET_STATE =                0x21
FHSS_START_SYNC =               0x22
FHSS_START_HOPPING =            0x23
FHSS_STOP_HOPPING =             0x24

FHSS_STATE_NONHOPPING =         0
FHSS_STATE_DISCOVERY =          1
FHSS_STATE_SYNCHING =           2
FHSS_LAST_NONHOPPING_STATE =    FHSS_STATE_SYNCHING

FHSS_STATE_SYNCHED =            3
FHSS_STATE_SYNC_MASTER =        4
FHSS_STATE_SYNCINGMASTER =      5
FHSS_LAST_STATE =               5       # used for testing


FHSS_STATES = {}
for key,val in list(globals().items()):
    if key.startswith("FHSS_STATE_"):
        FHSS_STATES[key] = val
        FHSS_STATES[val] = key
                
"""  MODULATIONS
Note that MSK is only supported for data rates above 26 kBaud and GFSK,
ASK , and OOK is only supported for data rate up until 250 kBaud. MSK
cannot be used if Manchester encoding/decoding is enabled.
"""
MOD_2FSK                        = 0x00
MOD_GFSK                        = 0x10
MOD_ASK_OOK                     = 0x30
MOD_4FSK                        = 0x40
MOD_MSK                         = 0x70
MANCHESTER                      = 0x08

MODULATIONS = {
        MOD_2FSK    : "2FSK",
        MOD_GFSK    : "GFSK",
        MOD_4FSK    : "4FSK",   # note: radio doesn't support Manchester encoding in 4FSK
        MOD_ASK_OOK : "ASK/OOK",
        MOD_MSK     : "MSK",
        MOD_2FSK | MANCHESTER    : "2FSK/Manchester encoding",
        MOD_GFSK | MANCHESTER    : "GFSK/Manchester encoding",
        MOD_ASK_OOK | MANCHESTER : "ASK/OOK/Manchester encoding",
        MOD_MSK  | MANCHESTER    : "MSK/Manchester encoding",
        }

SYNCMODES = {
        SYNCM_NONE: "None",
        SYNCM_15_of_16: "15 of 16 bits must match",
        SYNCM_16_of_16: "16 of 16 bits must match",
        SYNCM_30_of_32: "30 of 32 sync bits must match",
        SYNCM_CARRIER: "Carrier Detect",
        SYNCM_CARRIER_15_of_16: "Carrier Detect and 15 of 16 sync bits must match",
        SYNCM_CARRIER_16_of_16: "Carrier Detect and 16 of 16 sync bits must match",
        SYNCM_CARRIER_30_of_32: "Carrier Detect and 30 of 32 sync bits must match",
        }

BSLIMITS = {
        BSCFG_BS_LIMIT_0: "No data rate offset compensation performed",
        BSCFG_BS_LIMIT_3: "+/- 3.125% data rate offset",
        BSCFG_BS_LIMIT_6: "+/- 6.25% data rate offset",
        BSCFG_BS_LIMIT_12: "+/- 12.5% data rate offset",
        }

AESMODES = {
        ENCCS_MODE_CBC: "CBC - Cipher Block Chaining",
        ENCCS_MODE_CBCMAC: "CBC-MAC - Cipher Block Chaining Message Authentication Code",
        ENCCS_MODE_CFB: "CFB - Cipher Feedback",
        ENCCS_MODE_CTR: "CTR - Counter",
        ENCCS_MODE_ECB: "ECB - Electronic Codebook",
        ENCCS_MODE_OFB: "OFB - Output Feedback",
        }

NUM_PREAMBLE = [2, 3, 4, 6, 8, 12, 16, 24 ]

ADR_CHK_TYPES = [
        "No address check",
        "Address Check, No Broadcast",
        "Address Check, 0x00 is broadcast",
        "Address Check, 0x00 and 0xff are broadcast",
        ]



PKT_FORMATS = [
        "Normal mode",
        "reserved...",
        "Random TX mode",
        "reserved",
        ]

LENGTH_CONFIGS = [
        "Fixed Packet Mode",
        "Variable Packet Mode (len=first byte after sync word)",
        "reserved",
        "reserved",
        ]


# RFST (0xE1) - RF Strobe Commands
RFST_SFSTXON                    = 0x00
RFST_SCAL                       = 0x01
RFST_SRX                        = 0x02
RFST_STX                        = 0x03
RFST_SIDLE                      = 0x04
RFST_SNOP                       = 0x05

# 0xDF3B: MARCSTATE - Main Radio Control State Machine State
MARCSTATE_MARC_STATE            = 0x1F

MARC_STATE_SLEEP                = 0x00
MARC_STATE_IDLE                 = 0x01
MARC_STATE_VCOON_MC             = 0x03
MARC_STATE_REGON_MC             = 0x04
MARC_STATE_MANCAL               = 0x05
MARC_STATE_VCOON                = 0x06
MARC_STATE_REGON                = 0x07
MARC_STATE_STARTCAL             = 0x08
MARC_STATE_BWBOOST              = 0x09
MARC_STATE_FS_LOCK              = 0x0A
MARC_STATE_IFADCON              = 0x0B
MARC_STATE_ENDCAL               = 0x0C
MARC_STATE_RX                   = 0x0D
MARC_STATE_RX_END               = 0x0E
MARC_STATE_RX_RST               = 0x0F
MARC_STATE_TXRX_SWITCH          = 0x10
MARC_STATE_RX_OVERFLOW          = 0x11
MARC_STATE_FSTXON               = 0x12
MARC_STATE_TX                   = 0x13
MARC_STATE_TX_END               = 0x14
MARC_STATE_RXTX_SWITCH          = 0x15
MARC_STATE_TX_UNDERFLOW         = 0x16


MARC_STATE_MAPPINGS = [
    (0, 'MARC_STATE_SLEEP', RFST_SIDLE),
    (1, 'MARC_STATE_IDLE', RFST_SIDLE),
    (3, 'MARC_STATE_VCOON_MC', RFST_SIDLE),
    (4, 'MARC_STATE_REGON_MC', RFST_SIDLE),
    (5, 'MARC_STATE_MANCAL', RFST_SCAL),
    (6, 'MARC_STATE_VCOON', RFST_SIDLE),
    (7, 'MARC_STATE_REGON', RFST_SIDLE),
    (8, 'MARC_STATE_STARTCAL', RFST_SCAL),
    (9, 'MARC_STATE_BWBOOST', RFST_SIDLE),
    (10, 'MARC_STATE_FS_LOCK', RFST_SIDLE),
    (11, 'MARC_STATE_IFADCON', RFST_SIDLE),
    (12, 'MARC_STATE_ENDCAL', RFST_SCAL),
    (13, 'MARC_STATE_RX', RFST_SRX),
    (14, 'MARC_STATE_RX_END', RFST_SRX ),     # FIXME: this should actually be the config setting in register
    (15, 'MARC_STATE_RX_RST', RFST_SRX),
    (16, 'MARC_STATE_TXRX_SWITCH', RFST_SIDLE),
    (17, 'MARC_STATE_RX_OVERFLOW', RFST_SIDLE),
    (18, 'MARC_STATE_FSTXON', RFST_SFSTXON),
    (19, 'MARC_STATE_TX', RFST_STX),
    (20, 'MARC_STATE_TX_END', RFST_STX),        # FIXME: this should actually be the config setting in register
    (21, 'MARC_STATE_RXTX_SWITCH', RFST_SIDLE),
    (22, 'MARC_STATE_TX_UNDERFLOW', RFST_SIDLE) # FIXME: this should actually be the config setting in register
]

MODES = {}
for num,name,rfst in MARC_STATE_MAPPINGS:
    MODES[num] = name
    MODES[name] = num


T2SETTINGS = {}
T2SETTINGS_24MHz = {
    100: (4, 147, 3),
    150: (5, 110, 3),
    200: (5, 146, 3),
    250: (5, 183, 3),
    }
T2SETTINGS_26MHz = {
    100: (4, 158, 3),
    150: (5, 119, 3),
    200: (5, 158, 3),
    250: (5, 198, 3),
    }
    
TIP = (64,128,256,1024)

CHIPmhz = {
    0x91: 24,
    0x81: 26,
    0x11: 24,
    0x01: 26,
}

# FAKE DONGLE settings. mostly used for unittests.
FAKE_PARTNUM = 0x40
FAKE_DEBUG_CODES = (0x41, 0x42)
FAKE_DONGLE_BUILDDATA = b'DONS BROKEN FAKE DONGLE r0001\0'
FAKE_DONGLE_COMPILER = b'ATLASv100'
FAKE_DONGLE_SERIALNUM = b'0x47145'
FAKE_INTERRUPT_REGISTERS = {
        'IEN0': b'\xff',
        'IEN1': b'\x01',
        'IEN2': b'\xe0',
        'TCON': b'\xff',
        'S0CON': b'\x05',
        'IRCON': b'\x02',
        'IRCON2': b'\t',
        'S1CON': b'\x02',
        'RFIF': b'\x04',
        'DMAIE': b'\x01',
        'DMAIF': b'\x01',
        'DMAIRQ': b'\x10'}
# initial radio config
FAKE_MEM_DF00 = b'\x0cN\xff@\x00\x00\x00\x0c\x00%\x95U\xca\xa3\x01#\x116\x07\x0f\x18\x17l\x03@\x91\xb6\x10\xef*+\x1fY??\x881\t\x00\x00\x00\x00\x00\x00\x00\x00\xc0\x00\x00\x00\x00\x00\x00\x00\x11\x03\x12\x80\xaa\r\x90\xfd'



#!/usr/bin/env python
"""
#include<compiler.h>
/* ------------------------------------------------------------------------------------------------
*                                        Interrupt Vectors
* ------------------------------------------------------------------------------------------------
*/
#define  RFTXRX_VECTOR  0    /*  RF TX done / RX ready                       */
#define  ADC_VECTOR     1    /*  ADC End of Conversion                       */
#define  URX0_VECTOR    2    /*  USART0 RX Complete                          */
#define  URX1_VECTOR    3    /*  USART1 RX Complete                          */
#define  ENC_VECTOR     4    /*  AES Encryption/Decryption Complete          */
#define  ST_VECTOR      5    /*  Sleep Timer Compare                         */
#define  P2INT_VECTOR   6    /*  Port 2 Inputs                               */
#define  UTX0_VECTOR    7    /*  USART0 TX Complete                          */
#define  DMA_VECTOR     8    /*  DMA Transfer Complete                       */
#define  T1_VECTOR      9    /*  Timer 1 (16-bit) Capture/Compare/Overflow   */
#define  T2_VECTOR      10   /*  Timer 2 (MAC Timer) Overflow                */
#define  T3_VECTOR      11   /*  Timer 3 (8-bit) Capture/Compare/Overflow    */
#define  T4_VECTOR      12   /*  Timer 4 (8-bit) Capture/Compare/Overflow    */
#define  P0INT_VECTOR   13   /*  Port 0 Inputs                               */
#define  UTX1_VECTOR    14   /*  USART1 TX Complete                          */
#define  P1INT_VECTOR   15   /*  Port 1 Inputs                               */
#define  RF_VECTOR      16   /*  RF General Interrupts                       */
#define  WDT_VECTOR     17   /*  Watchdog Overflow in Timer Mode             */

SFR(P0,       0x80); // Port 0
  SBIT(P0_0,     0x80, 0); // Port 0 bit 0
  SBIT(P0_1,     0x80, 1); // Port 0 bit 1
  SBIT(P0_2,     0x80, 2); // Port 0 bit 2
  SBIT(P0_3,     0x80, 3); // Port 0 bit 3
  SBIT(P0_4,     0x80, 4); // Port 0 bit 4
  SBIT(P0_5,     0x80, 5); // Port 0 bit 5
  SBIT(P0_6,     0x80, 6); // Port 0 bit 6
  SBIT(P0_7,     0x80, 7); // Port 0 bit 7

SFR(SP,       0x81); // Stack Pointer
SFR(DPL0,     0x82); // Data Pointer 0 Low Byte
SFR(DPH0,     0x83); // Data Pointer 0 High Byte
SFR(DPL1,     0x84); // Data Pointer 1 Low Byte
SFR(DPH1,     0x85); // Data Pointer 1 High Byte
"""

from __future__ import print_function

import sys


def parseLines(lines):
    defs = {}
    incomment = False
    for line in lines:
        # find single-line comments
        slc = line.find("//")
        if (slc > -1):
            line = line[:slc]         + "#" + line[slc+2:]
        # find /* */ comments
        mlcs = line.find("/*")
        mlce = line.find("*/")
        if (mlcs>-1):
            if (mlce>-1):           # both are in this line
                if (mlce>mlcs):     # they are "together"
                    if (mlce >= len(line.strip())-3):
                        line = line[:mlcs] + '#' + line[mlcs+2:mlce]
                    else:
                        line = line[:mlcs] + '"""' + line[mlcs+2:mlce] + '"""' + line[mlce+2:]
                else:               # they are *not* together
                    line = line[mlce+2:mlcs]
            else:                   # only the beginning is in this line, treat like a single-line comment for now
                line = line[:mlcs]
                incomment = True
        elif incomment:              # no mlc-starter found... are we incomment?  then ignore until the end of comment
            if (mlce>-1):
                line = line[mlce+2:]
                incomment = False
            else:
                line = ''
        if incomment:                # if we're still incomment, this whole line is comment
            continue

        # chop initial and trailing whitespace
        line = line.strip()

        # now we can actually parse the line
        if (line.startswith("#define ")):
            line = line[8:].strip()     # peel off any additional spaces after the #define
            pieces = line.split(" ", 1)
            if len(pieces)<2:
                continue
            name, value = pieces
            if "(" in name:
                print(("SKIPPING: %s"%(line)), file=sys.stderr)
                continue                # skip adding "function" defines
            defs[name.strip()] = value.strip()
            
        elif (line.startswith("SFR(")):
            endparen = line.find(")")
            if (endparen == -1):
                print(("ERROR: SFR without end parens: '%s'"%(line)), file=sys.stderr)
                continue
            line = line[4:endparen].strip()
            name, value = line.split(",", 1)
            defs[name.strip()] = value.strip()
        elif (line.startswith("SFRX(")):
            endparen = line.find(")")
            if (endparen == -1):
                print(("ERROR: SFRX without end parens: '%s'"%(line)), file=sys.stderr)
                continue
            line = line[5:endparen].strip()
            name, value = line.split(",", 1)
            defs[name.strip()] = value.strip()
        elif (line.startswith("SBIT")):
            endparen = line.find(")")
            if (endparen == -1):
                print(("ERROR: SBIT without end parens: '%s'"%(line)), file=sys.stderr)
                continue
            line = line[5:endparen].strip()
            name, val1, val2 = line.split(",", 2)
            defs[name.strip()] = 1 << (int(val2.strip()))

    return defs


if __name__ == '__main__':
    defs = {}
    defs.update(parseLines(file('../includes/cc1110-ext.h')))
    defs.update(parseLines(file('../includes/cc1111.h')))
    defs.update(parseLines(file('/usr/share/sdcc/include/mcs51/cc1110.h')))

    skeys = list(defs.keys())
    skeys.sort()
    out = ["%-30s = %s"%(key,defs[key]) for key in skeys]

    trueout = []
    for x in out: 
        try:
            compile(x,'stdin','exec')
            trueout.append(x)
            print(x)
        except:
            sys.excepthook(*sys.exc_info())


            










import struct
try:
    from StringIO import StringIO ## for Python 2
except ImportError:
    from io import StringIO ## for Python 3

import vstruct.primitives as vs_prims

def isVstructType(x):
    return isinstance(x, vs_prims.v_base)

class VStruct(vs_prims.v_base):

    def __init__(self, bigend=False):
        # A tiny bit of evil...
        object.__setattr__(self, "_vs_values", {})
        vs_prims.v_base.__init__(self)
        self._vs_name = self.__class__.__name__
        self._vs_fields = []
        self._vs_field_align = False # To toggle visual studio style packing
        self._vs_padnum = 0
        self._vs_fmtbase = '<'
        if bigend:
            self._vs_fmtbase = '>'

    def vsGetClassPath(self):
        '''
        Return the entire class name (including module path).
        '''
        return '%s.%s' % (self.__module__, self._vs_name)

    def vsParse(self, bytes, offset=0):
        """
        For all the primitives contained within, allow them
        an opportunity to parse the given data and return the
        total offset...
        """
        plist = self.vsGetPrims()
        fmt = self.vsGetFormat()
        size = struct.calcsize(fmt)
        vals = struct.unpack(fmt, bytes[offset:offset+size])
        for i in range(len(plist)):
            plist[i].vsSetParsedValue(vals[i])

    def vsEmit(self):
        """
        Get back the byte sequence associated with this structure.
        """
        fmt = self.vsGetFormat()
        r = []
        for p in self.vsGetPrims():
            r.append(p.vsGetFmtValue())
        return struct.pack(fmt, *r)

    def vsGetFormat(self):
        """
        Return the format specifier which would then be used
        """
        # Unpack everything little endian, let vsParseValue deal...
        ret = self._vs_fmtbase
        for p in self.vsGetPrims():
            ret += p.vsGetFormat()
        return ret

    def vsIsPrim(self):
        return False

    def vsGetFields(self):
        ret = []
        for fname in self._vs_fields:
            fobj = self._vs_values.get(fname)
            ret.append((fname,fobj))
        return ret

    def vsGetField(self, name):
        x = self._vs_values.get(name)
        if x == None:
            raise Exception("Invalid field: %s" % name)
        return x

    def vsHasField(self, name):
        return self._vs_values.get(name) != None

    def vsSetField(self, name, value):
        if isVstructType(value):
            self._vs_values[name] = value
            return
        x = self._vs_values.get(name)
        return x.vsSetValue(value)

    # FIXME implement more arithmetic for structs...
    def __ixor__(self, other):
        for name,value in other._vs_values.items():
            self._vs_values[name] ^= value
        return self

    def vsAddField(self, name, value):
        if not isVstructType(value):
            raise Exception("Added fields MUST be vstruct types!")

        # Do optional field alignment...
        if self._vs_field_align:

            # If it's a primitive, all is well, if not, pad to size of
            # the first element of the VStruct/VArray...
            if value.vsIsPrim():
                align = len(value)
            else:
                fname = value._vs_fields[0]
                align = len(value._vs_values.get(fname))

            delta = len(self) % align
            if delta != 0:
                print("PADDING %s by %d" % (name,align-delta))
                pname = "_pad%d" % self._vs_padnum
                self._vs_padnum += 1
                self._vs_fields.append(pname)
                self._vs_values[pname] = vs_prims.v_bytes(align-delta)

        self._vs_fields.append(name)
        self._vs_values[name] = value

    def vsGetPrims(self):
        """
        return an order'd list of the primitive fields in this
        structure definition.  This is recursive and will return
        the sub fields of all nested structures.
        """
        ret = []
        for name, field in self.vsGetFields():
            if field.vsIsPrim():
                ret.append(field)
            else:
                ret.extend(field.vsGetPrims())

        return ret

    def vsGetTypeName(self):
        return self._vs_name

    def vsGetOffset(self, name):
        """
        Return the offset of a member.
        """
        offset = 0
        for fname in self._vs_fields:
            if name == fname:
                return offset
            x = self._vs_values.get(fname)
            offset += len(x)
        raise Exception("Invalid Field Specified!")

    def vsGetPrintInfo(self, offset=0, indent=0, top=True):
        ret = []
        if top:
            ret.append((offset, indent, self._vs_name, self))
        indent += 1
        for fname in self._vs_fields:
            x = self._vs_values.get(fname)
            off = offset + self.vsGetOffset(fname)
            if isinstance(x, VStruct):
                ret.append((off, indent, fname, x))
                ret.extend(x.vsGetPrintInfo(offset=off, indent=indent, top=False))
            else:
                ret.append((off, indent, fname, x))
        return ret

    def __len__(self):
        fmt = self.vsGetFormat()
        return struct.calcsize(fmt)

    def __getattr__(self, name):
        # Gotta do this for pickle issues...
        vsvals = self.__dict__.get("_vs_values")
        if vsvals == None:
            vsvals = {}
            self.__dict__["_vs_values"] = vsvals
        r = vsvals.get(name)
        if r is None:
            raise AttributeError(name)
        if isinstance(r, vs_prims.v_prim):
            return r.vsGetValue()
        return r

    def __setattr__(self, name, value):
        # If we have this field, asign to it
        x = self._vs_values.get(name, None)
        if x != None:
            return self.vsSetField(name, value)

        # If it's a vstruct type, create a new field
        if isVstructType(value):
            return self.vsAddField(name, value)

        # Fail over to standard object attribute behavior
        return object.__setattr__(self, name, value)

    def __iter__(self):
        # Our iteration returns name,field pairs
        ret = []
        for name in self._vs_fields:
            ret.append((name, self._vs_values.get(name)))
        return iter(ret)

    def __repr__(self):
        return self._vs_name

    def tree(self, va=0, reprmax=None):
        ret = ""
        for off, indent, name, field in self.vsGetPrintInfo():
            rstr = field.vsGetTypeName()
            if isinstance(field, vs_prims.v_number):
                val = field.vsGetValue()
                rstr = '0x%.8x (%d)' % (val,val)
            elif isinstance(field, vs_prims.v_prim):
                rstr = repr(field)
            if reprmax != None and len(rstr) > reprmax:
                rstr = rstr[:reprmax] + '...'
            ret += "%.8x (%.2d)%s %s: %s\n" % (va+off, len(field), " "*(indent*2),name,rstr)
        return ret

class VArray(VStruct):

    def __init__(self, elems=()):
        VStruct.__init__(self)
        for e in elems:
            self.vsAddElement(e)

    def vsAddElement(self, elem):
        """
        Used to add elements to an array
        """
        idx = len(self._vs_fields)
        self.vsAddField("%d" % idx, elem)

    def __getitem__(self, index):
        return self.vsGetField("%d" % index)

    #FIXME slice asignment

def resolve(impmod, nameparts):
    """
    Resolve the given (potentially nested) object
    from within a module.
    """
    if not nameparts:
        return None

    m = impmod
    for nname in nameparts:
        m = getattr(m, nname, None)
        if m == None:
            break

    return m

# NOTE: Gotta import this *after* VStruct/VSArray defined
import vstruct.defs as vs_defs

def getStructure(sname):
    """
    Return an instance of the specified structure.  The
    structure name may be a definition that was added with
    addStructure() or a python path (ie. win32.TEB) of a
    definition from within vstruct.defs.
    """
    x = resolve(vs_defs, sname.split("."))
    if x != None:
        return x()

    return None

def getModuleNames():
    return [x for x in dir(vs_defs) if not x.startswith("__")]

def getStructNames(modname):
    ret = []
    mod = resolve(vs_defs, modname)
    if mod == None:
        return ret

    for n in dir(mod):
        x = getattr(mod, n)
        if issubclass(x, VStruct):
            ret.append(n)

    return ret




from __future__ import print_function
from __future__ import division

from builtins import hex
from builtins import range
from builtins import bytes
from past.utils import old_div
import sys
import struct

fmtsLSB = [None, "B", "<H", "<I", "<I", "<Q", "<Q", "<Q", "<Q"]
fmtsMSB = [None, "B", ">H", ">I", ">I", ">Q", ">Q", ">Q", ">Q"]
sizes = [ 0, 1, 2, 4, 4, 8, 8, 8, 8]
masks = [ (1<<(8*i))-1 for i in range(9) ]

PYVER = int(sys.version[0])
def correctbytes(val):
    global PYVER

    if PYVER == 2:
        return chr(val)

    return bytes([val])

def wtfo(string):
    outstr = []
    bitlen = len(outstr) * 8
    for x in range(8):
        outstr.append(shiftString(string, x))

    string = strBitReverse(string)
    for x in range(8):
        outstr.append(shiftString(string, x))

    return outstr


def ord23(thing):
    if PYVER == 2:
        return ord(thing)
    return thing


def strBitReverse(string):
    # FIXME: this is really dependent upon python's number system.  large strings will not convert well.  
    # FIXME: break up array of 8-bit numbers and bit-swap in the array
    num = 0
    bits = len(string)*8
    # convert to MSB number
    for x in range(len(string)):
        ch = string[x]
        #num |= (ord(ch)<<(8*x))        # this is LSB
        num <<= 8
        num |= ord(ch)

    print(hex(num))
    rnum = bitReverse(num, bits)
    print(hex(rnum))

    # convert back from MSB number to string
    out = []
    for x in range(len(string)):
        out.append(correctbytes(rnum&0xff))
        rnum >>= 8
    out.reverse()
    print(''.join(out).encode('hex'))
    return ''.join(out)

def strXorMSB(string, xorval, size):
    '''
    lsb
    pads end of string with 00
    '''
    out = []
    strlen = len(string)
    string += "\x00" * sizes[size]

    for idx in range(0, strlen, size):
        tempstr = string[idx:idx+sizes[size]]
        temp, = struct.unpack( fmtsMSB[size], tempstr )
        temp ^= xorval
        temp &= masks[size]
        tempstr = struct.pack( fmtsMSB[size], temp )[-size:]
        out.append(tempstr)
    return ''.join(out)

        
        

def bitReverse(num, bitcnt):
    newnum = 0
    for idx in range(bitcnt):
        newnum <<= 1
        newnum |= num&1
        num    >>= 1
    return newnum

def shiftString(string, bits):
    carry = 0
    news = []
    for x in range(len(string)-1):
        newc = ((ord(string[x]) << bits) + (ord(string[x+1]) >> (8-bits))) & 0xff
        news.append("%c"%newc)
    newc = (ord(string[-1])<<bits) & 0xff
    news.append("%c"%newc)
    return "".join(news)

def getNextByte_feedbackRegister7bitsMSB():
    '''
    this returns a byte of a 7-bit feedback register stemming off bits 4 and 7
    the register is 7 bits long, but we return a more usable 8bits (ie. 
    '''
    global fbRegister

    retval = 0
    for x in range(8):      #MSB, 
        retval <<= 1
        retval |= (fbRegister >> 6)         # start with bit 7
        nb = ( ( fbRegister>>3) ^ (fbRegister>>6)) &1 
        fbRegister = ( ( fbRegister << 1 )   |   nb ) & 0x7f # do shifting
        #print("retval: %x  fbRegister: %x  bit7: %x  nb: %x" % (retval, fbRegister, (fbRegister>>6), nb))

    return retval

def getNextByte_feedbackRegister7bitsLSB():
    '''
    this returns a byte of a 7-bit feedback register stemming off bits 4 and 7
    the register is 7 bits long, but we return a more usable 8bits (ie. 
    '''
    global fbRegister

    retval = 0
    for x in range(8):      #MSB, 
        retval >>= 1
        retval |= ((fbRegister << 1)&0x80)         # start with bit 7

        nb = ( ( fbRegister>>3) ^ (fbRegister>>6)) &1 
        fbRegister = ( ( fbRegister << 1 )   |   nb ) & 0x7f # do shifting
        #print("retval: %x  fbRegister: %x  bit7: %x  nb: %x" % (retval, fbRegister, (fbRegister>>6), nb))

    return retval


def whitenData(data, seed=0xffff, getNextByte=getNextByte_feedbackRegister7bitsMSB):
    global fbRegister
    fbRegister = seed

    carry = 0
    news = []
    for x in range(len(data)-1):
        newc = ((ord(data[x]) ^ getNextByte() ) & 0xff)
        news.append("%c"%newc)
    return "".join(news)

def findSyncWord(byts, sensitivity=4, minpreamble=2): 
        '''
        seek SyncWords from a raw bitstream.  
        assumes we capture at least two (more likely 3 or more) preamble bytes
        '''
        possDwords = []
        # find the preamble (if any)
        while True:         # keep searching through string until we don't find any more preamble bits to pick on
            sbyts = byts
            pidx = byts.find("\xaa"*minpreamble)
            if pidx == -1:
                pidx = byts.find("\x55"*minpreamble)
                byts = shiftString(byts, 1)

            if pidx == -1:
                return possDwords
            
            # chop off the nonsense before the preamble
            sbyts = byts[pidx:]
            #print("sbyts: %s" % repr(sbyts))
            
            # find the definite end of the preamble (ie. it may be sooner, but we know this is the end)
            while (sbyts[0] == '\xaa' and len(sbyts)>2):
                sbyts = sbyts[1:]
            
            #print("sbyts: %s" % repr(sbyts))
            # now we look at the next 16 bits to narrow the possibilities to 8
            # at this point we have no hints at bit-alignment aside from 0xaa vs 0x55
            dwbits, = struct.unpack(">H", sbyts[:2])
            #print("sbyts: %s" % repr(sbyts))
            #print("dwbits: %s" % repr(dwbits))
            if len(sbyts)>=3:
                bitcnt = 0
                #  bits1 =      aaaaaaaaaaaaaaaabbbbbbbbbbbbbbbb
                #  bits2 =                      bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb
                bits1, = struct.unpack(">H", sbyts[:2])
                bits1 = bits1 | (ord('\xaa') << 16)
                bits1 = bits1 | (ord('\xaa') << 24)
                bits1 <<= 8
                bits1 |= (ord(sbyts[2]) )
                #print("bits: %x" % (bits1))

                bit = (5 * 8) - 2  # bytes times bits/byte          #FIXME: MAGIC NUMBERS!?
                while (bits1 & (3<<bit) == (2<<bit)):
                    bit -= 2
                #print("bit = %d" % bit)
                bits1 >>= (bit-16)
                #while (bits1 & 0x30000 != 0x20000): # now we align the end of the 101010 pattern with the beginning of the dword
                #    bits1 >>= 2
                #print("bits: %x" % (bits1))
                
                bitcount = min( 2 * sensitivity, 17 ) 
                for frontbits in range( bitcount ):            # with so many bit-inverted systems, let's not assume we know anything about the bit-arrangement.  \x55\x55 could be a perfectly reasonable preamble.
                    poss = (bits1 >> frontbits) & 0xffff
                    if not poss in possDwords:
                        possDwords.append(poss)
            byts = byts[pidx+1:]
        
        return possDwords

def findSyncWordDoubled(byts):
        possDwords = []
        # find the preamble (if any)
        bitoff = 0
        pidx = byts.find("\xaa\xaa")
        if pidx == -1:
            pidx = byts.find("\55\x55")
            bitoff = 1
        if pidx == -1:
            return []

        # chop off the nonsense before the preamble
        byts = byts[pidx:]

        # find the definite end of the preamble (ie. it may be sooner, but we know this is the end)
        while (byts[0] == ('\xaa', '\x55')[bitoff] and len(byts)>2):
            byts = byts[1:]

        # now we look at the next 16 bits to narrow the possibilities to 8
        # at this point we have no hints at bit-alignment
        dwbits, = struct.unpack(">H", byts[:2])
        if len(byts)>=5:
            bitcnt = 0
            #  bits1 =      aaaaaaaaaaaaaaaabbbbbbbbbbbbbbbb
            #  bits2 =                      bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb
            bits1, = struct.unpack(">H", byts[:2])
            bits1 = bits1 | (ord(('\xaa','\x55')[bitoff]) << 16)
            bits1 = bits1 | (ord(('\xaa','\x55')[bitoff]) << 24)
            bits1 <<= 8
            bits1 |= (ord(byts[2]) )
            bits1 >>= bitoff

            bits2, = struct.unpack(">L", byts[:4])
            bits2 <<= 8
            bits2 |= (ord(byts[4]) )
            bits2 >>= bitoff
            

            frontbits = 0
            for frontbits in range(16, 40, 2):    #FIXME: if this doesn't work, try 16, then 18+frontbits
                dwb1 = (bits1 >> (frontbits)) & 3
                dwb2 = (bits2 >> (frontbits)) & 3
                print("\tfrontbits: %d \t\t dwb1: %s dwb2: %s" % (frontbits, bin(bits1 >> (frontbits)), bin(bits2 >> (frontbits))))
                if dwb2 != dwb1:
                    break

            # frontbits now represents our unknowns...  let's go from the other side now
            for tailbits in range(16, -1, -2):
                dwb1 = (bits1 >> (tailbits)) & 3
                dwb2 = (bits2 >> (tailbits)) & 3
                print("\ttailbits: %d\t\t dwb1: %s dwb2: %s" % (tailbits, bin(bits1 >> (tailbits)), bin(bits2 >> (tailbits))))
                if dwb2 != dwb1:
                    tailbits += 2
                    break

            # now, if we have a double syncword, iinm, tailbits + frontbits >= 16
            print("frontbits: %d\t\t tailbits: %d, bits: %s " % (frontbits, tailbits, bin((bits2>>tailbits & 0xffffffff))))
            if (frontbits + tailbits >= 16):
                tbits = bits2 >> (tailbits&0xffff)
                tbits &= (0xffffffff)
                print("tbits: %x" % tbits)

                poss = tbits&0xffffffff
                if poss not in possDwords:
                    possDwords.append(poss)
            else:
                pass
                # FIXME: what if we *don't* have a double-sync word?  then we stop at AAblah or 55blah and take the next word?

            possDwords.reverse()
        return possDwords

#def test():

def visBits(data):
    pass



def getBit(data, bit):
    idx = old_div(bit, 8)
    bidx = bit % 8
    char = data[idx]
    return (ord(char)>>(7-bidx)) & 1



def detectRepeatPatterns(data, size=64, minEntropy=.07):
    #FIXME: convert strings to bit arrays before comparing.
    c1 = 0
    c2 = 0
    d1 = 0
    p1 = 0
    mask = (1<<size) - 1
    bitlen = 8*len(data)

    while p1 < (bitlen-size-8):
        d1 <<= 1
        d1 |= getBit(data, p1)
        d1 &= mask
        #print(bin(d1))

        if c1 < (size):
            p1 += 1
            c1 += 1
            continue

        d2 = 0
        p2 = p1+size
        while p2 < (bitlen):
            d2 <<= 1
            d2 |= getBit(data, p2)
            d2 &= mask
            #print(bin(d2))

            if c2 < (size):
                p2 += 1
                c2 += 1
                continue

            if d1 == d2 and d1 > 0:
                s1 = p1 - size
                s2 = p2 - size
                print("s1: %d\t  p1: %d\t  " % (s1, p1))
                print("s2: %d\t  p2: %d\t  " % (s2, p2))
                # complete the pattern until the numbers differ or meet
                while True:
                    p1 += 1
                    p2 += 1
                    #print("s1: %d\t  p1: %d\t  " % (s1, p1))
                    #print("s2: %d\t  p2: %d\t  " % (s2, p2))
                    if p2 >= bitlen:
                        break

                    b1 = getBit(data,p1)
                    b2 = getBit(data,p2)

                    if p1 == s2 or b1 != b2:
                        break

                length = p1 - s1
                c2 = 0
                p2 -= size

                bitSection, ent = bitSectString(data, s1, s1+length)
                if ent > minEntropy:
                    print("success:")
                    print("  * bit idx1: %4d (%4d bits) - '%s' %s" % (s1, length, bin(d1), bitSection.encode("hex")))
                    print("  * bit idx2: %4d (%4d bits) - '%s'" % (s2, length, bin(d2)))
            #else:
            #    print("  * idx1: %d - '%s'  * idx2: %d - '%s'" % (p1, d1, p2, d2))
            p2 += 1
        p1 += 1


def bitSectString(string, startbit, endbit):
    '''
    bitsects a string... ie. chops out the bits from the middle of the string
    returns the new string and the entropy (ratio of 0:1)
    '''
    ones = 0
    zeros = 0
    entropy = [zeros, ones]

    s = ''
    bit = startbit

    Bidx = old_div(bit, 8)
    bidx = (bit % 8)

    while bit < endbit:

        byte1 = ord( string[Bidx] )
        try:
            byte2 = ord( string[Bidx+1] )
        except IndexError:
            byte2 = 0

        byte = (byte1 << bidx) & 0xff
        byte |= (byte2 >> (8-bidx))
        #calculate entropy over the byte
        for bi in range(8):
            b = (byte>>bi) & 1
            entropy[b] += 1

        bit += 8
        Bidx += 1

        if bit > endbit:
            diff = bit-endbit
            mask = ~ ( (1<<diff) - 1 )
            byte &= mask

        s += correctbytes(byte)
    
    ent = old_div((min(entropy)+1.0), (max(entropy)+1))
    #print("entropy: %f" % ent)
    return (s, ent)


        
def genBitArray(string, startbit, endbit):
    '''
    bitsects a string... ie. chops out the bits from the middle of the string
    returns the new string and the entropy (ratio of 0:1)
    '''
    binStr, ent = bitSectString(string, startbit, endbit)

    s = []
    for byte in binStr:
        byte = ord(byte)
        for bitx in range(7, -1, -1):
            bit = (byte>>bitx) & 1
            s.append(bit)

    return (s, ent)


chars_top = [
        " ", #000
        " ", #001
        "^", #010
        "/", #011
        " ", #100
        " ", #101
        "\\",#110
        "-", #111
        ]

chars_mid = [
        " ", #000
        "|", #001
        "#", #010
        " ", #011
        "|", #100
        "#", #101
        " ", #110
        " ", #110
        ]

chars_bot = [
        "-", #000
        "/", #001
        " ", #010
        " ", #011
        "\\",#100
        "V", #101
        " ", #110
        " ", #110
        ]


def reprBitArray(bitAry, width=194):
    top = []
    mid = []
    bot = []

    arylen = len(bitAry)
    # top line
    #FIXME: UGGGGLY and kinda broken.
    fraction = 1.0 * arylen/width
    expand = [bitAry[int(x*fraction)] for x in range(width)]

    for bindex in range(width):
        bits = 0
        if bindex>0:
            bits += (expand[bindex-1]) << (2)
        bits += (expand[bindex]) << (1)
        if bindex < width-1:
            bits += (expand[bindex+1])

        top.append( chars_top[ bits ] )
        mid.append( chars_mid[ bits ] )
        bot.append( chars_bot[ bits ] )

    tops = "".join(top)
    mids = "".join(mid)
    bots = "".join(bot)
    return "\n".join([tops, mids, bots])

def invertBits(data):
    output = []
    ldata = len(data)
    off = 0

    if ldata&1:
        output.append( correctbytes( ord( data[0] ) ^ 0xff) )
        off = 1

    if ldata&2:
        output.append( struct.pack( "<H", struct.unpack( "<H", data[off:off+2] )[0] ^ 0xffff) )
        off += 2

    #method 1
    #for idx in xrange( off, ldata, 4):
    #    output.append( struct.pack( "<I", struct.unpack( "<I", data[idx:idx+4] )[0] & 0xffff) )

    #method2
    count = old_div(ldata, 4)
    #print(ldata, count)
    numlist = struct.unpack( "<%dI" % count, data[off:] )
    modlist = [ struct.pack("<L", (x^0xffffffff) ) for x in numlist ]
    output.extend(modlist)

    return ''.join(output)


def diff_manchester_decode(data, align=False):
    # FIXME: not validated.  must find reference to validate against
    '''
    differential manchester encoding/decoding uses 2 symbols per data bit.
    there must always be a transition between the first and second symbol of a bit.
    bit values are determined by the existence/lack of transition between symbol pairs

    set align=True to allow *one* sync of the bits to clock.  ie, either the whole 
    thing lines up with a transition in the middle of every bit, or shift one, and
    try again.  one way *must* have transitions, or failure occurs
    '''
    syncd = False

    out = []
    last = 0
    obyte = 0
    for bidx in range(len(data)):
        byte = ord(data[bidx])
        for y in range(6, -1, -2):
            if not syncd:
                diff = last & 1
                bit0 = (byte >> (y+1)) & 1
                bit1 = (byte >> y) & 1
            else:
                diff = last >> 1
                bit0 = last & 1
                bit1 = (byte >> (y+1)) & 1

            if bit0 == bit1:
                if syncd or not align:
                    raise Exception("Differential Manchester Decoder cannot work with this data.  Sync fault at index %d,%d" % (bidx, y))

                syncd = 1
                # redo the last stuff with new info
                diff = last >> 1
                bit0 = last & 1
                bit1 = (byte >> (y+1)) & 1

            obyte <<= 1
            if diff != bit0:
                obyte |= 1

            last = (bit0 << 1) | bit1
        if (bidx & 1): 
            out.append(correctbytes(obyte))
            obyte = 0

    if not (bidx & 1):
        obyte << 4 # pad 0's on end
        out.append(correctbytes(obyte))
    return ''.join(out)



def biphase_mark_coding_encode(data):
    # FIXME: broken?  this looks more like BMC (biphase mark encoding)
    # FIXME: write encoder as well
    out = []
    last = 0
    for bidx in range(len(data)):
        byte = ord(data[bidx])
        obyte = 0
        for y in range(7, -1, -1):
            bit = (byte >> y) & 1
            obyte <<= 1
            if bit == last:
                obyte |= 1

            last = bit
        if bidx & 1:
            print("%d - write" % bidx)
            out.append(correctbytes(obyte))
        else:
            print("%d - skip" % bidx)
    if not (bidx & 1):
        print("%d - write" % bidx)
        out.append(correctbytes(obyte))

    return ''.join(out)

def manchester_decode(data, hilo=1):
    out = []
    last = 0
    obyte = 0
    for bidx in range(len(data)):
        byte = ord(data[bidx])
        for y in range(7, -1, -1):
            bit = (byte >> y) & 1

            if not (y & 1):   # every other bit counts
                obyte <<= 1
                if bit and not last:
                    if not hilo:
                        obyte |= 1
                elif last and not bit:
                    if hilo:
                        obyte |= 1

            last = bit
        if (bidx & 1): 
            out.append(correctbytes(obyte))
            obyte = 0

    if not (bidx & 1):
        obyte << 4 # pad 0's on end
        out.append(correctbytes(obyte))
    return ''.join(out)

def manchester_encode(data, hilo=1):
    '''
    for the sake of testing.
    assumings msb, and 
    '''
    if hilo:
        bits = (0b01, 0b10)
    else:
        bits = (0b10, 0b01)

    out = []
    for bidx in range(len(data)):
        byte = ord(data[bidx])
        obyte = 0
        for bitx in range(7,-1,-1):
            bit = (byte>>bitx) & 1
            obyte <<= 2
            obyte |= bits[bit]

        out.append(struct.pack(">H", obyte))
    return ''.join(out)

def findManchesterData(data, hilo=1):
    poss = []

    for x in range(8):
        try:
            newdata = shiftString(data, x)
            thing = manchester_decode(x, newdata)
            poss.append(thing)
        except:
            pass

def findManchester(data, minbytes=10):
    print("DEBUG: DATA=" + repr(data))
    success = []

    last = 0
    last2 = 0
    lastCount = 0
    minbits = minbytes * 8

    for bidx in range(len(data)):
        byt = ord(data[bidx])
        for btidx in range(0, 8, 2):
            # compare every other bits
            bit = (byt>>(8-btidx)) & 1

            if (bit + last + last2) in (1,2):
                lastCount += 1
            else:
                # we're done, or not started
                if lastCount >= minbits:
                    lenbytes = (old_div(lastCount, 8))
                    lenbits = lastCount % 8
                    startbyte = bidx - lenbytes
                    if lenbits > btidx:
                        startbyte += 1
                        lenbits -= 8
                    startbit = btidx - lenbits

                    stopbyte = startbyte + lenbytes + (0,1)[lenbits>0]
                    bytez = data[startbyte:stopbyte]

                    success.append((bidx, startbyte, startbit, bytez))
                lastCount = 0
            # cycle through
            last2 = last
            last = bit
    return success




import rflib

import usb
import time
import queue
import logging
import unittest
import threading
import traceback

from rflib.const import *
from rflib.bits import ord23

logging.basicConfig(level=logging.INFO, format='%(asctime)s:%(levelname)s:%(name)s: %(message)s')
logger = logging.getLogger(__name__)

EP0BUFSIZE = 512


class fakeMemory:
    def __init__(self, size=64*1024):
        self.memory = [0 for x in range(size)]
        self.mmio = {}
        #self.mmio[X_RFST] = self.mmio_RFST
        #self.mmio[X_RFST] = self.mmio_MARCSTATE

    def readMemory(self, addr, size):
        logger.debug("fm.readMemory(0x%x, 0x%x)", addr, size)
        chunk = b''.join([b'%c' % x for x in self.memory[addr:addr+size]])
        if len(chunk) < size:
            chunk += b"@" * (size-len(chunk))
        return chunk

    def writeMemory(self, addr, data):
        logger.debug("fm.writeMemory(0x%x, %r)", addr, data)
        #if type(data) == str:
        #    raise(Exception("Cannot write 'str' to fakeMemory!  Must use 'bytes'"))

        for x in range(len(data)):
            tgt = addr+x
            val = data[x]

            handler = self.mmio.get(tgt)
            if handler is not None:
                val = handler(tgt, data[x])

            # if we didn't return None from the handler, write it anyway
            if val is not None:
                self.memory[tgt] = val

    def mmio_RFST(self, tgt, dbyte):
        logger.info('mmio_RFST(0x%x, %r)', tgt, dbyte)
        print("RFST==%x  (%x)" % (self.readMemory(X_RFST, 1), ord(dbyte)))


        # configure MARCSTATE
        val = ord(dbyte)
        if val in (2, 3):
            val = dbyte+10

        else:
            val = MARC_STATE_RX

        self.writeMemory(MARCSTATE, b'%c'%(val))

        # still set RFST
        return dbyte

    def mmio_MARCSTATE(self, tgt, dbyte):
        rfst = self.readMemory(X_RFST, 1)
        logger.info('mmio_MARCSTATE(0x%x, %r) rfst=%r', tgt, dbyte, rfst)
        return MARC_STATE_RX

class fakeDon:
    pass



MAX_CHANNELS            =   880
MAX_TX_MSGS             =   2
MAX_TX_MSGLEN           =   240   # must match RF_MAX_TX_CHUNK in rflib/chipcon_nic.py
                                  # and be divisible by 16 for crypto operations
DEFAULT_NUM_CHANS       = 83
DEFAULT_NUM_CHANHOPS    = 83

class MAC_Data:
    def __init__(self):
        self.mac_state = FHSS_STATE_NONHOPPING
        self.MAC_threshold = 6              # when the T2 clock as overflowed this many times, change channel
        self.MAC_timer = 0                  # this tracks how many times it's overflowed (really?  32-bits for these two?!?)
        self.NumChannels = DEFAULT_NUM_CHANS                # in case of multiple paths through the available channels 
        self.NumChannelHops = DEFAULT_NUM_CHANHOPS             # total number of channels in pattern (>= g_MaxChannels)
        self.curChanIdx = 0                 # indicates current channel index of the hopping pattern
        self.tLastStateChange = 0
        self.tLastHop = 0
        self.desperatelySeeking = 0         # this should be unnecessary, and should instead use mac_state?
        self.txMsgIdx = 0
        self.txMsgIdxDone = 0
        self.synched_chans = 0

    def serialize(self):
        return struct.pack("<B8H2BH", 
               self.mac_state,
               self.MAC_threshold,
               self.MAC_timer,
               self.NumChannels,
               self.NumChannelHops,
               self.curChanIdx,
               self.tLastStateChange,
               self.tLastHop,
               self.desperatelySeeking,
               self.txMsgIdx,
               self.txMsgIdxDone,
               self.synched_chans)

    def deserialize(self, data):
       (self.mac_state,
               self.MAC_threshold,
               self.MAC_timer,
               self.NumChannels,
               self.NumChannelHops,
               self.curChanIdx,
               self.tLastStateChange,
               self.tLastHop,
               self.desperatelySeeking,
               self.txMsgIdx,
               self.txMsgIdxDone,
               self.synched_chans) = struct.unpack("<B8H2BH", data)


class fakeDongle:
    '''
    This class emulates a real RfCat dongle (the physical device), as well as LibUSB.
    '''
    def __init__(self):
        self._recvbuf = b''
        self.bulk5 = queue.Queue()
        self.bulk0 = [0 for x in range(EP0BUFSIZE)]
        self.memory = fakeMemory()

        self.start_ts = time.time()
        self.aesMode = 0
        self.ampMode = 0
        self.macdata = MAC_Data()
        self.NIC_ID = 0
        self.g_txMsgQueue = ['\0'*(MAX_TX_MSGLEN+1) for x in range(MAX_TX_MSGS)]
        self.g_Channels = b''

        self.memory.writeMemory(0xdf00, FAKE_MEM_DF00)
        self.memory.writeMemory(0xdf46, b'\xf0\x0d')
        for intreg, intval in FAKE_INTERRUPT_REGISTERS.items():
            logger.info('setting interrupt register: %r = %r', intreg, intval)
            self.memory.writeMemory(eval(intreg), intval)

    def clock(self):
        return time.time() - self.start_ts

    def controlMsg(self, flags, request, buf, value, index, timeout):
        logger.info("controlMsg: 0x%x %r %r 0x%x %r %r", flags, request, buf, value, index, timeout)
        try:
            # split by direction (IN/OUT)
            if flags & USB_BM_REQTYPE_DIR_IN:
                logger.info("<= fakeDoer.controlMsg(flags=0x%x, request=%r, buf=%r, value=%r, index=%x, timeout=%r)", flags, request, buf, value, index, timeout)
                if request == EP0_CMD_GET_DEBUG_CODES:
                    return b'AB'
                if request == EP0_CMD_PEEKX:
                    return self.memory.readMemory(value, buf)

            else:  # flags & USB_BM_REQTYPE_DIR_OUT fails since USB_BM_REQTYPE_DIR_OUT == 0!
                logger.info("=> fakeDoer.controlMsg(flags=0x%x, request=%r, buf=%r, value=%r, index=%x, timeout=%r)", flags, request, buf, value, index, timeout)
                if request == EP0_CMD_POKEX:
                    self.memory.writeMemory(value, buf)

        except:
            logger.error(traceback.format_exc())

    def txdata(self, app, cmd, data):
        if type(data) == int and data < 0x100:
            data = b'%c' % data
        self.bulk5.put(struct.pack('<BBH', app, cmd, len(data)) + data)

    def bulkWrite(self, chan, buf, timeout=1):
        try:
            # handle write "parts"
            buflen = len(buf)   # need to return this, because that's what the libusb interface does.
            self._recvbuf += buf
            logger.debug("=> fakeDoer.bulkWrite(5, %r)", buf)

            curbuflen = len(self._recvbuf)
            if curbuflen < 4:
                return buflen

            app, cmd, mlen = struct.unpack("<BBH", self._recvbuf[:4])

            if curbuflen < mlen+2:
                logger.info("bulkWrite: returning because buffer isn't big enough: len: %x  need: %x", curbuflen, mlen+2)
                return buflen

            # now handle a packet
            pkt = self._recvbuf[:mlen+4]

            data = pkt[4:]
            #print("_recvbuf:%r\t\tpkt:%r\t\tapp:%x\tcmd:%x\tdata:%r\t\tmlen:%r\t" % (self._recvbuf, pkt, app, cmd, data, hex(mlen)))
            self._recvbuf = self._recvbuf[mlen+4:]

            # handle commands for the SYSTEM app
            if app == APP_SYSTEM:
                if cmd == SYS_CMD_PEEK:
                    size, addr = struct.unpack("<HH", data[:4])
                    retmsg = self.memory.readMemory(addr, size)
                    self.txdata(app, cmd, retmsg) 

                elif cmd == SYS_CMD_POKE:
                    addr, = struct.unpack("<H", data[:2])
                    size = mlen - 2
                    chunk = data[2:2+size]
                    logger.info("=>> POKE: pkt:%r\t\tdata:%r\t\tsize:%r\t\taddr:%r\t\t%r", repr(pkt), repr(data), hex(size), hex(addr), chunk)
                    self.memory.writeMemory(addr, chunk)

                    self.bulk5.put(pkt)

                elif cmd == SYS_CMD_PING:
                    self.bulk5.put(pkt)

                elif cmd == SYS_CMD_BUILDTYPE:
                    self.txdata(app, cmd, FAKE_DONGLE_BUILDDATA)

                elif cmd == SYS_CMD_COMPILER:
                    self.txdata(app, cmd, FAKE_DONGLE_COMPILER)

                elif cmd == SYS_CMD_DEVICE_SERIAL_NUMBER:
                    self.txdata(app, cmd, FAKE_DONGLE_SERIALNUM)

                elif cmd == SYS_CMD_RFMODE:
                    if len(data) > 1:
                        logger.warning("ummm. what's this extra data in your SYS_CMD_RFMODE command?")
                    if len(data) == 0:
                        logger.warning("SYS_CMD_RFMODE: need a byte to put in X_RFST!")
                    else:
                        self.memory.writeMemory(X_RFST, data[0:1])
                    self.txdata(app, cmd, data)

                else:
                    self.log(b'WTFO!  no APP_SYSTEM::0x%x', cmd)
                    self.bulk5.put(pkt)

            # handle commands for the NIC app
            elif app == APP_NIC:
                if cmd == NIC_GET_AES_MODE:
                    self.txdata(app, cmd, b'%c' % self.aesMode)

                elif cmd == NIC_SET_AES_MODE:
                    self.aesMode = ord23(data[0])
                    self.txdata(app, cmd, b'%c' % self.aesMode)

                elif cmd == NIC_SET_AMP_MODE:
                    self.ampMode = ord23(data[0])
                    self.txdata(app, cmd, b'%c' % self.ampMode)

                elif cmd == NIC_GET_AMP_MODE:
                    self.txdata(app, cmd, b'%c' % self.ampMode)

                elif cmd == NIC_SET_AES_IV:
                    self.setAES(data, ENCCS_CMD_LDIV, (self.aesMode & AES_CRYPTO_MODE))
                    self.txdata(app, cmd, data[:16])

                elif cmd == NIC_SET_AES_KEY:
                    self.setAES(data, ENCCS_CMD_LDKEY, (self.aesMode & AES_CRYPTO_MODE))
                    self.txdata(app, cmd, data[:16])

                elif cmd == NIC_SET_ID:
                    # fixme: sending 8 bit to 16 bit function???
                    self.NIC_ID = ord23(data[0])
                    self.txdata(app, cmd, data[0])

                elif cmd == NIC_LONG_XMIT:
                    # load up macdata queues, follow-on with 
                    #
                    #
                    # this is duplicating our work in transmit_long().  pick one.
                    if (macdata.mac_state != FHSS_STATE_NONHOPPING):
                        data[0] = RC_RF_MODE_INCOMPAT
                        self.txdata(app, cmd, data[0])
                   
                    else:
                        length, blocks = struct.unpack("<HB", data[:2])
                        txTotal= 0
                        data[0] = transmit_long(data[3:], length, blocks)
                        self.txdata(app, cmd, data[0])

                elif cmd == NIC_LONG_XMIT_MORE:
                    length = ord23(data[0])
                    if (length == 0):
                        if(rfTxTotalTXLen):
                            self.debug("dropout final wait!")
                            #debughex16(rfTxTotalTXLen)
                            #debughex(g_txMsgQueue[0][0])
                            #debughex(g_txMsgQueue[1][0])
                            self.lastCode[1] = LCE_DROPPED_PACKET
                            data[0] = RC_TX_DROPPED_PACKET
                            #LED = 0
                            #resetRFSTATE()
                            self.macdata.mac_state = FHSS_STATE_NONHOPPING
                            self.txdata(app, cmd, b'%c' % RC_TX_DROPPED_PACKET)
                            return
                        
                        #LED = 0
                        self.macdata.mac_state = FHSS_STATE_NONHOPPING
                        self.debug("total bytes tx:")
                        #debughex16(txTotal)
                        self.txdata(app, cmd, b'%c' % LCE_NO_ERROR)
                        return
                    
                    # catch if we've been called out of sequence, or we've had an underrun
                    if (self.macdata.mac_state != FHSS_STATE_LONG_XMIT):
                        self.debug("underrun")
                        # TX underrun
                        if(self.lastCode[1] == LCE_DROPPED_PACKET):
                            self.txdata(app, cmd, b'%c' % RC_TX_DROPPED_PACKET)
                            
                        else:
                            self.lastCode[1] = LCE_RF_MULTI_BUFFER_NOT_INIT
                            self.txdata(app, cmd, b'%c' % RC_RF_MODE_INCOMPAT)
                        
                        #LED = 0
                        #resetRFSTATE()
                        self.macdata.mac_state = FHSS_STATE_NONHOPPING
                    else:
                        # add data to rolling datafer
                        #data[0] = MAC_tx(&data[1], (__xdata u8) len)
                        # check for any other error return
                        #if(data[0] && data[0] != RC_ERR_BUFFER_NOT_AVAILABLE)
                        #{
                        #    debug("datafer error");
                        #    debughex(data[0]);
                        #    LED = 0;
                        #    resetRFSTATE();
                        #    self.macdata.mac_state = FHSS_STATE_NONHOPPING;
                        #}
                        self.txdata(app, cmd, data[0]);

                elif cmd == FHSS_XMIT:
                    length = ord23(data[0])
                    #len += (*data++) << 8;
                    #repeat = *data++;
                    #repeat += (*data++) << 8;
                    #offset = *data++;
                    #offset += (*data++) << 8;
                    #transmit(data, len, repeat, offset);
                    #MAC_tx(data, len);
                    ##/// for some strange reason, if we call this in MAC_tx it dies, but not from here. ugh.
                    if (length > MAX_TX_MSGLEN):
                        self.debug("FHSSxmit message too long");
                        self.txdata(app, cmd, b'%c' % length);
                        return buflen

                    elif (self.g_txMsgQueue[self.macdata.txMsgIdx][0] != 0):
                        self.debug("still waiting on the last packet");
                        self.txdata(app, cmd, b'%c' % length);
                        return buflen

                    g_txMsgQueue[self.macdata.txMsgIdx][0] = length
                    g_txMsgQueue[self.macdata.txMsgIdx][1] = data[1:]

                    self.macdata.txMsgIdx += 1
                    if (self.macdata.txMsgIdx >= MAX_TX_MSGS):
                        self.macdata.txMsgIdx = 0;

                    self.txdata(app, cmd,  b'%c' % length)
                    
                elif cmd == FHSS_SET_CHANNELS:
                    self.macdata.NumChannels = ord23(data[0])
                    if (self.macdata.NumChannels <= MAX_CHANNELS):
                        self.g_Channels = data[2:self.macdata.NumChannels]
                        self.txdata(app, cmd, struct.pack("<H", self.macdata.NumChannels))

                    else:
                        self.txdata(app, cmd, b"NO DEAL")

                elif cmd == FHSS_GET_CHANNELS:
                    self.txdata(app, cmd, self.g_Channels)

                elif cmd == FHSS_NEXT_CHANNEL:
                    #MAC_set_chanidx(MAC_getNextChannel());
                    self.macdata.curChanIdx += 1

                    chan = self.setFHSSchanByIdx(self.macdata.curChanIdx)
                    self.txdata(app, cmd, b'%c' % chan) 

                elif cmd == FHSS_CHANGE_CHANNEL:
                    #PHY_set_channel(data[0]);
                    self.memory.writeMemory(CHANNR, data[0])
                    self.txdata(app, cmd, data[0]);

                elif cmd == FHSS_START_HOPPING:
                    self.begin_hopping(0);
                    self.txdata(app, cmd, data[0]);

                elif cmd == FHSS_STOP_HOPPING:
                    self.stop_hopping();
                    self.txdata(app, cmd, data[0]);

                elif cmd == FHSS_SET_MAC_THRESHOLD:
                    self.macdata.MAC_threshold = ord23(data[0])
                    self.txdata(app, cmd, data[0]);

                elif cmd == FHSS_GET_MAC_THRESHOLD:
                    self.txdata(app, cmd, struct.pack("<I", self.macdata.MAC_threshold))

                elif cmd == FHSS_SET_MAC_DATA:
                    self.debugx(data);
                    #debughex(data[0]);
                    self.macdata.deserialize(data)
                    self.txdata(app, cmd, data);

                elif cmd == FHSS_GET_MAC_DATA:
                    self.macdata.MAC_timer = self.get_rf_MAC_timer()
                    self.txdata(app, cmd, self.macdata.serialize());

                elif cmd == FHSS_START_SYNC:
                    #MAC_sync(data[0])
                    self.txdata(app, cmd, data[0]);
                    
                elif cmd == FHSS_SET_STATE:
                    # store the main timer value for beginning of this phase.
                    self.macdata.tLastStateChange = self.clock()
                    self.macdata.mac_state = ord23(data[0])
                    
                    # if macdata.mac_state is > 2, make sure the T2 interrupt is set
                    # if macdata.mac_state <= 2, make sure T2 interrupt is ignored
                    if self.macdata.mac_state in (FHSS_STATE_NONHOPPING, FHSS_STATE_DISCOVERY, FHSS_STATE_SYNCHING):
                        self.stop_hopping();

                    elif self.macdata.mac_state == FHSS_STATE_SYNCINGMASTER:
                        self.MAC_do_Master_scanny_thingy();

                    elif self.macdata.mac_state in (FHSS_STATE_SYNCHED, FHSS_STATE_SYNC_MASTER):
                        self.begin_hopping(0);
                    
                    self.txdata(app, cmd, data[0]);
                    
                elif cmd == FHSS_GET_STATE:
                    self.txdata(app, cmd, self.macdata.mac_state)
                    
                else:
                    self.log(b'WTFO!  no APP_NIC::0x%x', cmd)
                    self.bulk5.put(pkt)
            else:
                # everything else...  just echo
                self.bulk5.put(pkt)

            return buflen

        except:
            logger.error(traceback.format_exc())

    def bulkRead(self, chan, length, timeout=1):
        '''
        In standard USB fashion, bulkRead() handles the "IN" communication, whereby the "host" 
        pulls information back from the "device".  ie. our responses to commands.

        RfCat polls this function repeatedly, to provide the illusion of bi-directional
        communication, when in fact USB (pre-v3) is completely host-driven.  If a USB device
        gets to talk, it's because the host asked for information.

        For our purposes, bulkRead() simply pops data out of the EP5 Bulk "queue" and returns.

        This has *nothing* to do with "reading" from the memory.  bulkRead() gives the dongle
        the "talking stick"
        '''
        starttime = time.time()

        while time.time() - starttime < timeout:
            try:
                out = self.bulk5.get_nowait()
                logger.debug('<= fakeDoer.bulkRead(5, %r) == %r', length, out)
                return b"@" + out
            except queue.Empty:
                time.sleep(.05)

            logger.debug('<= fakeDoer.bulkRead(5, %r) == <EmptyQueue>', length)
            raise usb.USBError('Operation timed out (FakeDongle)')

    def log(self, msg, *args):
        if len(args):
            msg = msg % args
        self.txdata(APP_DEBUG, DEBUG_CMD_STRING, msg)

    # no need to reinvent the wheel, this is a difference in CC1111 memories, not functionality
    debug  = log
    debugx = log

    def setAES(self, data, cmd, flags):
        '''
        For now, we do nothing.
        '''
        return

    def setFHSSchanByIdx(self, chanidx):
        chan = self.g_Channels[chanidx]
        self.memory.writeMemory(CHANNR, chan)
        return chan

    def begin_hopping(self, startchan):
        self.memory.writeMemory(CHANNR, b'%c' % startchan)
        return
    def stop_hopping(self):
        return

    def get_rf_MAC_timer(self):
        return int((self.clock() * 20) % self.macdata.MAC_threshold)

class FakeRfCat(rflib.RfCat):
    def __init__(self, idx=0, debug=False, copyDongle=None, RfMode=RFST_SRX):
        # instantiate ourself as an official RfCat dongle
        rflib.RfCat.__init__(self, idx, debug, copyDongle, RfMode)

    def _internal_select_dongle(self, console=False):
        self._d = fakeDon()
        self._do = fakeDongle()
        self.console = console

    def getPartNum(self):
        return FAKE_PARTNUM




#!/usr/bin/env python
from __future__ import print_function
import sys

sn_header = """// Serial number
               10,                      // bLength
               USB_DESC_STRING,         // bDescriptorType
"""

try:
    f = open('.serial', 'r')
    ser = int(f.read(), 16) #+ 1
    f.close()
except IOError:
    ser = 0

print("[--- new serial number: %.4x ---]" % ser, file=sys.stderr)

f = open('.serial', 'w')
f.write("%.4x" % ser)
f.close()

sertxt = "%.4x" % ser

for c in sertxt:
    n = ord(c)
    sys.stdout.write("%s,0," % n)




'''
Structure definitions for the OSX MachO binary format.
'''

import vstruct
import vstruct.primitives as vs_prim

from vstruct.defs.macho.const import *

from vstruct.defs.macho.fat import *
from vstruct.defs.macho.loader import *




import os
import tempfile
import unittest
from rflib.const import *
from rflib.fakedongle_nic import FakeRfCat


testhex = ''':10000000020102FFFFFFFFFFFFFFFFFFFFFFFFFFF8
:10001000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0
:10002000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFE0
:10003000FFFFFF022010FFFFFFFFFFFFFFFFFFFF9B
:10004000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFC0
:10005000FFFFFF02164EFFFFFFFFFFFFFFFFFFFF47
:10006000FFFFFFFFFFFFFFFFFFFFFF022065FFFF16
:10007000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF90
:10FF4000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFC1
:10FF5000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFB1
:10FF6000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFA1
:10FF7000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF91
:10FF8000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF81
:10FF9000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF71
:10FFA000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF61
:10FFB000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF51
:10FFC000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF41
:10FFD000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF31
:10FFE000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF21
:10FFF000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF11
:00000001FF
'''


d = FakeRfCat()

class TestApis(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        global d
        unittest.TestCase.__init__(self, *args, **kwargs)
        self.d = d

    def test_api_usb(self):
        self.assertEqual(self.d.getPartNum(), FAKE_PARTNUM)
        self.assertEqual(self.d.getDebugCodes(), FAKE_DEBUG_CODES)
        self.assertEqual(self.d.getBuildInfo(), FAKE_DONGLE_BUILDDATA)
        self.assertEqual(self.d.getCompilerInfo(), FAKE_DONGLE_COMPILER)
        self.assertEqual(self.d.getDeviceSerialNumber(), FAKE_DONGLE_SERIALNUM)
        self.assertEqual(self.d.getInterruptRegisters(), FAKE_INTERRUPT_REGISTERS)
        
        '''
        rflib/chipcon_usb.py:106:    def setRFparameters(self):
        '''

    def test_api_nic(self):
        self.assertEqual(self.d.getRadioConfig(), FAKE_MEM_DF00)
        #self.d.printRadioConfig()
        self.d.setRadioConfig(bytedef=b'@'*0x3a)
        self.assertEqual(self.d.getRadioConfig(), b'@'*0x3a + b'\xaa\r\x90\xfd')
        self.d.setRadioConfig(bytedef=FAKE_MEM_DF00)
        self.d.setModeRX()
        self.d.poke(X_RFST, b'%c'%RFST_SRX)
        self.assertEqual(self.d.getRadioConfig()[:-5], FAKE_MEM_DF00[:-5])
        self.d.printRadioConfig()

        self.d.setRfMode(RFST_SRX)
        self.assertEqual(ord(self.d.peek(X_RFST)), RFST_SRX)
        self.d.setModeTX()
        self.assertEqual(ord(self.d.peek(X_RFST)), RFST_STX)
        self.d.setModeRX()
        self.assertEqual(ord(self.d.peek(X_RFST)), RFST_SRX)
        self.d.setModeIDLE()
        self.assertEqual(ord(self.d.peek(X_RFST)), RFST_SIDLE)

        self.d.setLedMode(1)
        self.d.setLedMode(0)
        self.d.getMARCSTATE()

        self.d.setEnableCCA()
        
        self.d.setFreq(878e6)
        freq, freqnum = self.d.getFreq()
        self.assertAlmostEqual(freq, 878e6, delta=200)
        self.d.setPower(0xc0)
        self.d.setMaxPower()

        self.assertEqual(self.d.getPktLEN(), (255,0))
        self.d.setPktPQT(3)
        self.assertEqual(self.d.getPktPQT(), 3)
        self.d.setMdmModulation(MOD_4FSK)
        self.assertEqual(self.d.getMdmModulation(), MOD_4FSK)
        
        self.d.setMdmChanSpc(chanspc=333e3)
        self.assertAlmostEqual(self.d.getMdmChanSpc(), 333e3, delta=300)

        self.d.setEnablePktCRC()
        self.assertEqual(self.d.getEnablePktCRC(), True)

        self.d.setEnablePktDataWhitening()
        self.assertEqual(self.d.getEnablePktDataWhitening(), True)

        self.d.setEnablePktAppendStatus()
        self.assertEqual(self.d.getEnablePktAppendStatus(), True)

        self.d.setEnableMdmManchester()
        self.assertEqual(self.d.getEnableMdmManchester(), True)

        self.d.setEnableMdmFEC()
        self.assertEqual(self.d.getEnableMdmFEC(), True)

        self.d.setEnableMdmDCFilter()
        self.assertEqual(self.d.getEnableMdmDCFilter(), True)

        self.d.setFsIF(freq_if=23.5e3)
        self.assertAlmostEqual(self.d.getFsIF(), 23.5e3, delta=80)

        self.d.setAmpMode(ampmode=1)
        self.assertEqual(self.d.getAmpMode(), 1)

        self.d.setPktAddr(addr=4)
        self.assertEqual(ord(self.d.getPktAddr()), 4)

        self.d.setFsOffset(if_off=4, mhz=24, radiocfg=None)
        self.assertEqual(self.d.getFsOffset(mhz=24, radiocfg=None), 4)

        self.d.setChannel(channr=0x40, radiocfg=None)
        self.assertEqual(self.d.getChannel(radiocfg=None), 0x40)

        self.d.setMdmChanBW(bw=93.7e3, mhz=24, radiocfg=None)
        self.assertAlmostEqual(self.d.getMdmChanBW(mhz=24, radiocfg=None), 93.7e3, delta=100)

        self.d.setMdmDRate(drate=57.6e3, mhz=24, radiocfg=None)
        self.assertAlmostEqual(self.d.getMdmDRate(mhz=24, radiocfg=None), 57.6e3, delta=100)

        self.d.setMdmDeviatn(deviatn=20.5e3, mhz=24, radiocfg=None)
        self.assertAlmostEqual(self.d.getMdmDeviatn(mhz=24, radiocfg=None), 20.5e3, delta=100)

        self.d.setMdmSyncWord(word=0x7145, radiocfg=None)
        self.assertEqual(self.d.getMdmSyncWord(radiocfg=None), 0x7145)
        
        self.d.setMdmSyncMode(syncmode=SYNCM_15_of_16, radiocfg=None)
        self.assertEqual(self.d.getMdmSyncMode(radiocfg=None), SYNCM_15_of_16)

        self.d.setMdmNumPreamble(preamble=MFMCFG1_NUM_PREAMBLE_4, radiocfg=None)
        self.assertEqual(self.d.getMdmNumPreamble(radiocfg=None), MFMCFG1_NUM_PREAMBLE_4)
        
        self.d.setBSLimit(bslimit=3, radiocfg=None)
        self.assertEqual(self.d.getBSLimit(radiocfg=None), 3)

        self.d.getRSSI()
        self.d.getLQI()

        self.d.setAESmode(aesmode=AES_CRYPTO_DEFAULT)
        self.d.getAESmode()
        self.d.setAESiv(iv= b'@'*16)
        self.d.setAESkey(key= b'@'*16)

        self.d.setChannels(channels=[1,1,2,3,5,8,13,21,34,55,89,144])
        self.d.getChannels()



        self.d.setMACperiod(dwell_ms=20, mhz=24)

        fakemacdata = (0, 6, 0, 83, 83, 0, 0, 0, 0, 0, 0, 0)
        self.d.setMACdata(fakemacdata)
        testmacdata = self.d.getMACdata()
        self.assertEqual(fakemacdata[3:], testmacdata[3:])
        self.assertEqual(fakemacdata[0:2],  testmacdata[0:2])
        self.assertAlmostEqual(fakemacdata[2],  testmacdata[2], delta=10)     # MAC timer, ticks 20x per second

        self.d.setMACthreshold(value=24)
        self.assertEqual(self.d.getMACthreshold(), 24)

        self.d.setFHSSstate(state=FHSS_STATE_SYNCHED)
        self.assertEqual(self.d.getFHSSstate(), ('FHSS_STATE_SYNCHED', 3))
        '''
        self.d.setRFbits(addr, bitnum, bitsz, val, suppress=False)
        self.d.setEnDeCoder(endec=None)
        self.d.getValueFromReprString(stringarray, line_text)
        '''


    def test_bits(self):
        import rflib.bits as rfbits
        

    def test_intelhex(self):
        import rflib.intelhex as rfhex
        temphexfn = os.sep.join([tempfile.gettempdir(), 'rfcat_unittest_tempfile.hex'])
        # write the file first
        #open(temphexfn, 'w').write(testhex)
        with open(temphexfn, 'w') as f:
            f.write(testhex)

        ih = rfhex.IntelHex()
        ih.loadhex(temphexfn)



'''
rflib/bits.py:108:def getNextByte_feedbackRegister7bitsMSB():
rflib/bits.py:125:def getNextByte_feedbackRegister7bitsLSB():
rflib/bits.py:294:def getBit(data, bit):
rflib/__init__.py:184:    def setMdmSyncWord(self, word, radiocfg=None):

'''

'''  should figure out some unittest for these guys.
rfcat_msfrelay:70:    def set_freq(self, args):
rfcat_msfrelay:79:    def get_modulations(self):
rfcat_msfrelay:83:    def set_modulation(self, args):
rfcat_msfrelay:118:    def set_mode(self, args):
rfcat_msfrelay:140:    def set_channel(self, args):
rfcat_msfrelay:146:    def set_channel_bandwidth(self, args):
rfcat_msfrelay:155:    def set_channel_spc(self, args):
rfcat_msfrelay:172:    def set_baud_rate(self, args):
rfcat_msfrelay:184:    def set_deviation(self, args):
rfcat_msfrelay:196:    def set_sync_word(self, args):
rfcat_msfrelay:202:    def set_sync_mode(self, args):
rfcat_msfrelay:208:    def set_number_preamble(self, args):
rfcat_msfrelay:214:    def set_lowball(self):
rfcat_msfrelay:218:    def set_maxpower(self):
rfcat_msfrelay:222:    def set_power(self, args):
rfcat_server:815:    def getInterruptRegisters(self):
rfcat_server:819:    def getMACthreshold(self):
rfcat_server:820:    def setMACthreshold(self, value):
rfcat_server:821:    def setFHSSstate(self, state):
rfcat_server:822:    def getFHSSstate(self):
'''



#!/usr/bin/env python2

from __future__ import print_function

import sys
from rflib.intelhex import IntelHex

WRITEBACK = True
if len(sys.argv) > 1:
    WRITEBACK = False
    ser = int(sys.argv[1])
else:
    try:
        ser = int(file(".serial", 'rb').read(), 16) + 1
    except IOError:
        ser = 0

print(("[--- new serial number: %.4x ---]" % ser), file=sys.stderr)

if WRITEBACK:
    file(".serial", 'wb').write("%.13x" % ser)

sertxt = ''
sertmp = "%.13x" % ser
for c in sertmp:
    sertxt += "%s\x00" % c

ihc=IntelHex('CCBootloader/CCBootloader-rfcat-chronosdongle.hex')
ihd=IntelHex('CCBootloader/CCBootloader-rfcat-donsdongle.hex')
ihy=IntelHex('CCBootloader/CCBootloader-rfcat-ys1.hex')
ihc.puts(0x13e0, "@las\x1c\x03" + sertxt)
ihd.puts(0x13e0, "@las\x1c\x03" + sertxt)
ihy.puts(0x13e0, "@las\x1c\x03" + sertxt)
ihc.write_hex_file('CCBootloader/CCBootloader-rfcat-chronosdongle-serial.hex')
ihd.write_hex_file('CCBootloader/CCBootloader-rfcat-donsdongle-serial.hex')
ihy.write_hex_file('CCBootloader/CCBootloader-rfcat-ys1-serial.hex')




#!/usr/bin/python

from __future__ import print_function

import sys
import serial

port = "ACM0"
if len(sys.argv) > 1:
    port = sys.argv.pop()

dport = "/dev/tty" + port

print("Opening serial port %s for listening..." % dport)
s=serial.Serial(dport, 115200)

counter = 0
while True:
    print("%d: %s" % (counter, repr(s.read(12))))
    counter += 1
    #sys.stdout.write(s.read(1))
    #sys.stdout.flush()



# firmware errors
LCE_NO_ERROR                            = 0x00
LCE_USB_EP5_TX_WHILE_INBUF_WRITTEN      = 0x01
LCE_USB_EP0_SENT_STALL                  = 0x04
LCE_USB_EP5_OUT_WHILE_OUTBUF_WRITTEN    = 0x05
LCE_USB_EP5_LEN_TOO_BIG                 = 0x06
LCE_USB_EP5_GOT_CRAP                    = 0x07
LCE_USB_EP5_STALL                       = 0x08
LCE_USB_DATA_LEFTOVER_FLAGS             = 0x09

LCE_RF_RXOVF                            = 0x10
LCE_RF_TXUNF                            = 0x11
LCE_DROPPED_PACKET                      = 0x12
LCE_RFTX_NEVER_TX                       = 0x13
LCE_RFTX_NEVER_LEAVE_TX                 = 0x14
LCE_RF_MODE_INCOMPAT                    = 0x15
LCE_RF_BLOCKSIZE_INCOMPAT               = 0x16
LCE_RF_MULTI_BUFFER_NOT_INIT            = 0x17
LCE_RF_MULTI_BUFFER_NOT_FREE            = 0x18

RC_NO_ERROR                             = 0x00
RC_TX_DROPPED_PACKET                    = 0xec
RC_TX_ERROR                             = 0xed
RC_RF_BLOCKSIZE_INCOMPAT                = 0xee
RC_RF_MODE_INCOMPAT                     = 0xef
RC_TEMP_ERR_BUFFER_NOT_AVAILABLE        = 0xfe # temporary error - retry!
RC_ERR_BUFFER_SIZE_EXCEEDED             = 0xff
RC_FAIL_TRANSMIT_LONG                   = 0xffff

# python client only errors
PY_NO_ERROR                             = 0x00
PY_TX_BLOCKSIZE_INCOMPAT                = 0xd0 # block > 255 bytes but with with repeat or offset specified
PY_TX_BLOCKSIZE_TOO_LARGE               = 0xda # block > 65535 (we could make this bigger by using a u32 for size)




import vstruct
import vstruct.primitives as vs_prim

class fat_header(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self, bigend=True)
        self.magic = vs_prim.v_uint32()
        self.nfat_arch = vs_prim.v_uint32()

class fat_arch(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self, bigend=True)
        self.cputype    = vs_prim.v_uint32()  # cpu specifier (int) */
        self.cpusubtype = vs_prim.v_uint32()  # machine specifier (int) */
        self.offset     = vs_prim.v_uint32()  # file offset to this object file */
        self.size       = vs_prim.v_uint32()  # size of this object file */
        self.align      = vs_prim.v_uint32()  # alignment as a power of 2 */




#!/usr/bin/env ipython3

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from builtins import bytes
from builtins import str
from builtins import hex
from builtins import range
from builtins import object
from past.utils import old_div
import os
import sys
import usb
import time
import struct
import select
import threading
from binascii import hexlify

from . import bits
from .bits import correctbytes, ord23
from .const import *

if os.name == 'nt':
    import msvcrt

def keystop(delay=0):
    if os.name == 'posix':
        return len(select.select([sys.stdin],[],[],delay)[0])
    else:
        return msvcrt.kbhit()

def getRfCatDevices():
    '''
    returns a list of USB device objects for any rfcats that are plugged in
    NOTE: if any rfcats are in bootloader mode, this will cause python to Exit
    '''
    rfcats = []
    for bus in usb.busses():
        for dev in bus.devices:
            # OpenMoko assigned or Legacy TI
            if (dev.idVendor == 0x0451 and dev.idProduct == 0x4715) or (dev.idVendor == 0x1d50 and (dev.idProduct == 0x6047 or dev.idProduct == 0x6048 or dev.idProduct == 0x605b or dev.idProduct == 0xecc1)):
                rfcats.append(dev)

            elif (dev.idVendor == 0x1d50 and (dev.idProduct == 0x6049 or dev.idProduct == 0x604a or dev.idProduct == 0xecc0)):
                print("Already in Bootloader Mode... exiting")
                exit(0)

    return rfcats

class ChipconUsbTimeoutException(Exception):
    def __str__(self):
        return "Timeout waiting for USB response."

direct=False

class USBDongle(object):
    ######## INITIALIZATION ########
    def __init__(self, idx=0, debug=False, copyDongle=None, RfMode=RFST_SRX):
        self.chipnum = None
        self.chipstr = "uninitialized"
        self.rsema = None
        self.xsema = None
        self._bootloader = False
        self._init_on_reconnect = True
        self._do = None
        self.idx = idx
        self.cleanup()
        self._debug = debug
        self._quiet = False
        self._threadGo = threading.Event()
        self._recv_time = 0
        self.radiocfg = RadioConfig()
        self._rfmode = RfMode
        self._radio_configured = False

        self.ctrl_thread = threading.Thread(target=self.run_ctrl)
        self.ctrl_thread.setDaemon(True)
        self.ctrl_thread.start()

        self.recv_thread = threading.Thread(target=self.runEP5_recv)
        self.recv_thread.setDaemon(True)
        self.recv_thread.start()

        self.send_thread = threading.Thread(target=self.runEP5_send)
        self.send_thread.setDaemon(True)
        self.send_thread.start()

        self.resetup(copyDongle=copyDongle)
        self.max_packet_size = USB_MAX_BLOCK_SIZE

    def cleanup(self):
        self._usberrorcnt = 0;
        self.recv_queue = b''
        self.recv_mbox  = {}
        self.recv_event = threading.Event()
        self.xmit_event = threading.Event()
        self.reset_event = threading.Event()
        self.xmit_queue = []
        self.xmit_event.clear()
        self.reset_event.clear()
        self.trash = []
   
    def setRFparameters(self):
        pass

    def run_ctrl(self):
        '''
        we wait for reset events and run resetup
        '''
        while True:
            self.reset_event.wait()
            self.resetup(False)
            self.reset_event.clear()
            time.sleep(4)

    def setup(self, console=True, copyDongle=None):
        global dongles

        if copyDongle is not None:
            self.devnum = copyDongle.devnum
            self.chipnum = copyDongle.chipnum
            self.chipstr = copyDongle.chipstr
            self._d = copyDongle._d
            self._do = copyDongle._do
            self._usbmaxi = copyDongle._usbmaxi
            self._usbmaxo = copyDongle._usbmaxo
            self._usbcfg = copyDongle._usbcfg
            self._usbintf = copyDongle._usbintf
            self._usbeps = copyDongle._usbeps
            self._threadGo.set()
            self.ep5timeout = EP_TIMEOUT_ACTIVE
            copyDongle._threadGo.clear()            # we're taking over from here.
            self.rsema = copyDongle.rsema
            self.xsema = copyDongle.xsema
            return

        self._internal_select_dongle(console)
        self.finish_setup()

    def _internal_select_dongle(self, console=False):
        '''
        strap in USB interface.  this has to insert a ._d and ._do widget with 
        the correct USB-like interface
        '''
        self.ep5timeout = EP_TIMEOUT_ACTIVE

        dongles = []
        for dev in getRfCatDevices():
            if self._debug: print((dev), file=sys.stderr)
            do = dev.open()
            iSN = do.getDescriptor(1,0,50)[16]
            devnum = dev.devnum
            dongles.append((devnum, dev, do))

        dongles.sort()
        if len(dongles) == 0:
            raise Exception("No Dongle Found.  Please insert a RFCAT dongle.")

        # claim that interface!
        do = dongles[self.idx][2]
        
        try:
            do.claimInterface(0)
        except Exception as e:
            if console or self._debug: print(("Error claiming usb interface:" + repr(e)), file=sys.stderr)



        self.devnum, self._d, self._do = dongles[self.idx]

        self._usbcfg = self._d.configurations[0]
        self._usbintf = self._usbcfg.interfaces[0][0]
        self._usbeps = self._usbintf.endpoints
        for ep in self._usbeps:
            if ep.address & 0x80:
                self._usbmaxi = ep.maxPacketSize
            else:
                self._usbmaxo = ep.maxPacketSize

    def finish_setup(self):
        '''
        we've finished selecting and strapping in the usb dongle interface... continue
        '''
        self.rsema = threading.Lock()
        self.xsema = threading.Lock()

        self._usbmaxi, self._usbmaxo = (EP5IN_MAX_PACKET_SIZE, EP5OUT_MAX_PACKET_SIZE)
        self._threadGo.set()

        self.getRadioConfig()
        chip = self.getPartNum()
        chipstr = CHIPS.get(chip)

        self.chipnum = chip
        self.chipstr = chipstr

        if chip == None:
            print("Older firmware, consider upgrading.")
        else:
            self.chipstr = "unrecognized dongle: %s" % chip

        if self._init_on_reconnect:
            if self._radio_configured:
                self._clear_buffers()
                self.setRadioConfig()
            else:
                self.setRFparameters()
                self._radio_configured = True

    def resetup(self, console=True, copyDongle=None):
        self._do=None
        if self._bootloader: 
            return
        if self._debug: print(("waiting (resetup) %x" % self.idx), file=sys.stderr)
        while (self._do==None):
            try:
                self.setup(console, copyDongle)
                if copyDongle is None:
                    self._clear_buffers(False)
                self.ping(3, wait=10, silent=True)
                self.setRfMode(self._rfmode)

            except Exception as e:
                #if console: sys.stderr.write('.')
                if not self._quiet:
                    print(("Error in resetup():" + repr(e)), file=sys.stderr)
                #if console or self._debug: print("Error in resetup():" + repr(e), file=sys.stderr)
                time.sleep(1)


    ########  BASE FOUNDATIONAL "HIDDEN" CALLS ########
    def _sendEP0(self, request=0, buf=None, value=0x200, index=0, timeout=DEFAULT_USB_TIMEOUT):
        if buf == None:
            buf = b'HELLO THERE'
        return self._do.controlMsg(USB_BM_REQTYPE_TGT_EP|USB_BM_REQTYPE_TYPE_VENDOR|USB_BM_REQTYPE_DIR_OUT, request, buf, value, index, timeout), buf

    def _recvEP0(self, request=0, length=64, value=0, index=0, timeout=100):
        retary = [b"%c"%x for x in self._do.controlMsg(USB_BM_REQTYPE_TGT_EP|USB_BM_REQTYPE_TYPE_VENDOR|USB_BM_REQTYPE_DIR_IN, request, length, value, index, timeout)]
        if len(retary):
            return b''.join(retary)
        return b""

    def _sendEP5(self, buf=None, timeout=DEFAULT_USB_TIMEOUT):
        global direct
        if (buf==None):
            buf = b"\xff\x82\x07\x00ABCDEFG"
        if direct:
            self._do.bulkWrite(5, buf, timeout)
            return

        while (len(buf)>0):
            drain = buf[:self._usbmaxo]
            buf = buf[self._usbmaxo:]

            if self._debug: print("XMIT:"+repr(drain), file=sys.stderr)
            try:
                numwrt = self._do.bulkWrite(5, drain, timeout)
                if numwrt != len(drain):
                    raise Exception("Didn't write all the data!? Sent: %d != Queued: %d.  REqueuing!(this may be the wrong thing to do, swat me if so)" % (numwrt, len(drain)))
            except Exception as e:
                if self._debug: print("requeuing on error '%s' (%s)" % (repr(drain), e), file=sys.stderr)
                self.xsema.acquire()
                msg = self.xmit_queue.insert(0, drain)
                self.xmit_event.set()
                self.xsema.release()
                if self._debug: print(repr(self.xmit_queue), file=sys.stderr)
        
    def _recvEP5(self, timeout=100):
        retary = [b"%c"%x for x in self._do.bulkRead(0x85, 500, timeout)]
        if self._debug: print("RECV:"+repr(retary), file=sys.stderr)
        if len(retary):
            return b''.join(retary)
        return b''

    def _clear_buffers(self, clear_recv_mbox=False):
        threadGoSet = self._threadGo.isSet()
        self._threadGo.clear()
        if self._debug:
            print(("_clear_buffers()"), file=sys.stderr)
        if clear_recv_mbox:
            for key in list(self.recv_mbox.keys()):
                self.trash.extend(self.recvAll(key))
        elif self.recv_mbox.get(APP_SYSTEM) != None:
            self.trash.extend(self.recvAll(APP_SYSTEM))
        self.trash.append((time.time(),self.recv_queue))
        self.recv_queue = b''
        # self.xmit_queue = []          # do we want to keep this?
        if threadGoSet: self._threadGo.set()


    ######## TRANSMIT/RECEIVE THREADING ########
    def runEP5_send(self):
        msg = b''
        self.send_threadcounter = 0

        while True:
            self._threadGo.wait()
            self.send_threadcounter = (self.send_threadcounter + 1) & 0xffffffff

            #### transmit stuff.  if any exists in the xmit_queue
            self.xmit_event.wait() # event driven xmit
            msgsent = False

            try:
                if len(self.xmit_queue):
                    self.xsema.acquire()

                    msg = self.xmit_queue.pop(0)
                    if not len(self.xmit_queue): # if there was only one message
                        self.xmit_event.clear() # clear the queue, within the lock
                    
                    self.xsema.release()

                    self._sendEP5(msg)
                    msgsent = True

                else:
                    if self._debug>3: sys.stderr.write("NoMsgToSend ")
            except:
                sys.excepthook(*sys.exc_info())

    def runEP5_recv(self):
        msg = b''
        self.recv_threadcounter = 0

        while True:
            self._threadGo.wait()
            if self._debug>3: sys.stderr.write(".")

            self.recv_threadcounter = (self.recv_threadcounter + 1) & 0xffffffff
            msgrecv = False

            #### handle debug application
            try:
                q = None
                b = self.recv_mbox.get(APP_DEBUG, None)
                if (b != None):
                    for cmd in list(b.keys()):
                        q = b[cmd]
                        if len(q):
                            buf,timestamp = q.pop(0)
                            if self._debug > 1: print(("recvthread: buf length: %x\t\t cmd: %x\t\t(%s)"%(len(buf), cmd, repr(buf))), file=sys.stderr)

                            if (cmd == DEBUG_CMD_STRING):
                                if (len(buf) < 4):
                                    if (len(q)):
                                        buf2 = q.pop(0)
                                        buf += buf2
                                    q.insert(0,buf)
                                    if self._debug: sys.stderr.write('*')
                                else:
                                    length, = struct.unpack("<H", buf[2:4])
                                    if self._debug >1: print(("len=%d"%length), file=sys.stderr)
                                    if (len(buf) < 4+length):
                                        if (len(q)):
                                            buf2 = q.pop(0)
                                            buf += buf2
                                        q.insert(0,buf)
                                        if self._debug: sys.stderr.write('&')
                                    else:
                                        printbuf = buf[4:4+length]
                                        requeuebuf = buf[4+length:]
                                        if len(requeuebuf):
                                            if self._debug>1:  print((" - DEBUG..requeuing %s"%repr(requeuebuf)), file=sys.stderr)
                                            q.insert(0,requeuebuf)
                                        print(("DEBUG: (%.3f) %s" % (timestamp, repr(printbuf))), file=sys.stderr)
                            elif (cmd == DEBUG_CMD_HEX):
                                #print(repr(buf), file=sys.stderr)
                                print("DEBUG: (%.3f) 0x%x %d"%(timestamp, struct.unpack("B", buf[4:5])[0], struct.unpack("B", buf[4:5])[0]), file=sys.stderr)
                            elif (cmd == DEBUG_CMD_HEX16):
                                #print(repr(buf), file=sys.stderr)
                                print("DEBUG: (%.3f) 0x%x %d"%(timestamp, struct.unpack("<H", buf[4:6])[0], struct.unpack("<H", buf[4:6])[0]), file=sys.stderr)
                            elif (cmd == DEBUG_CMD_HEX32):
                                #print(repr(buf), file=sys.stderr)
                                print("DEBUG: (%.3f) 0x%x %d"%(timestamp, struct.unpack("<L", buf[4:8])[0], struct.unpack("<L", buf[4:8])[0]), file=sys.stderr)
                            elif (cmd == DEBUG_CMD_INT):
                                print("DEBUG: (%.3f) %d"%(timestamp, struct.unpack("<L", buf[4:8])[0]), file=sys.stderr)
                            else:
                                print(('DEBUG COMMAND UNKNOWN: %x (buf=%s)'%(cmd,repr(buf))), file=sys.stderr)

            except:
                sys.excepthook(*sys.exc_info())

            #### receive stuff.
            if self._debug>2: print("recvthread: Doing receiving...",self.ep5timeout, file=sys.stderr)
            try:
                #### first we populate the queue
                msg = self._recvEP5(timeout=self.ep5timeout)
                if len(msg) > 0:
                    self.recv_queue += msg
                    msgrecv = True
            except usb.USBError as e:
                #sys.stderr.write(repr(self.recv_queue))
                #sys.stderr.write(repr(e))
                errstr = repr(e)
                if self._debug>4: print(repr(sys.exc_info()), file=sys.stderr)
                if ('No error' in errstr):
                    pass
                elif ('Connection timed out' in errstr):
                    pass
                elif ('Operation timed out' in errstr):
                    pass
                else:
                    if ('could not release intf' in errstr):
                        if self._debug: print("skipping")
                        pass
                    elif ('No such device' in errstr):
                        self._threadGo.clear()
                        #self.resetup(False) ## THIS IS A PROBLEM.
                        self.reset_event.set()
                        print("===== RESETUP set from recv thread")
                    elif ('Input/output error' in errstr):  # USBerror 5
                        self._threadGo.clear()
                        #self.resetup(False) ## THIS IS A PROBLEM.
                        self.reset_event.set()
                        print("===== RESETUP set from recv thread")

                    else:
                        if self._debug: print("Error in runEP5() (receiving): %s" % errstr)
                        if self._debug>2: sys.excepthook(*sys.exc_info())
                    self._usberrorcnt += 1
                pass
            except AttributeError as e:
                if "'NoneType' object has no attribute 'bInterfaceNumber'" in str(e):
                    print("Error: dongle went away.  USB bus problems?")
                    self._threadGo.clear()
                    #self.resetup(False)
                    self.reset_event.set()

            except:
                sys.excepthook(*sys.exc_info())

            if self._debug>2: print("recvthread: Sorting mail...", file=sys.stderr)
            #### parse, sort, and deliver the mail.
            try:
                # FIXME: is this robust?  or just overcomplex?
                if len(self.recv_queue):
                    idx = self.recv_queue.find(b'@')
                    if (idx==-1):
                        if self._debug > 3:
                            sys.stderr.write('@')
                    else:
                        if (idx>0):
                            if self._debug: print(("runEP5(): idx>0?"), file=sys.stderr)
                            self.trash.append(self.recv_queue[:idx])
                            self.recv_queue = self.recv_queue[idx:]
                   
                        # recv_queue is vulnerable here, but it's ok because we only modify it earlier in this same thread
                        # DON'T CHANGE recv_queue from other threads!
                        msg = self.recv_queue
                        msglen = len(msg)
                        #if self._debug > 2: print( "Sorting msg", len(msg), hexlify(msg)
                        while (msglen>=5):                                      # if not enough to parse length... we'll wait.
                            if not self._recv_time:                             # should be 0 to start and when done with a packet
                                self._recv_time = time.time()
                            app = ord23(msg[1])
                            cmd = ord23(msg[2])
                            length, = struct.unpack("<H", msg[3:5])

                            if self._debug>1: print(("recvthread: app=%x  cmd=%x  len=%x"%(app,cmd,length)), file=sys.stderr)

                            if (msglen >= length+5):
                                #### if the queue has enough characters to handle the next message... chop it and put it in the appropriate recv_mbox
                                msg = self.recv_queue[1:length+5]                   # drop the initial '@' and chop out the right number of chars
                                self.recv_queue = self.recv_queue[length+5:]        # chop it out of the queue

                                b = self.recv_mbox.get(app,None)

                                if self.rsema.acquire():                            # THREAD SAFETY DANCE
                                    try:
                                        if (b == None):
                                            b = {}
                                            self.recv_mbox[app] = b
                                    except:
                                        sys.excepthook(*sys.exc_info())
                                    finally:
                                        self.rsema.release()                            # THREAD SAFETY DANCE COMPLETE
                               
                                q = b.get(cmd)

                                if self.rsema.acquire():                            # THREAD SAFETY DANCE
                                    try:
                                        if (q is None):
                                            q = []
                                            b[cmd] = q

                                        q.append((msg, self._recv_time))

                                        # notify receivers that a new msg is available
                                        self.recv_event.set()
                                        self._recv_time = 0                         # we've delivered the current message

                                    except:
                                        sys.excepthook(*sys.exc_info())
                                    finally:
                                        self.rsema.release()                            # THREAD SAFETY DANCE COMPLETE
                               
                            else:            
                                if self._debug>1:     sys.stderr.write('=')

                            msg = self.recv_queue
                            msglen = len(msg)
                            # end of while loop

            except:
                sys.excepthook(*sys.exc_info())

            if self._debug>2: print("readthread: Loop finished", file=sys.stderr)
            if not (msgrecv or len(msg)) :
                #time.sleep(.1)
                self.ep5timeout = EP_TIMEOUT_IDLE
            else:
                self.ep5timeout = EP_TIMEOUT_ACTIVE
                if self._debug > 5:  sys.stderr.write(" %s:%d .-P."%(msgrecv,len(msg)))



    ######## APPLICATION API ########
    def recv(self, app, cmd=None, wait=USB_RX_WAIT):
        '''
        high-level USB EP5 receive.  
        checks the mbox for app "app" and command "cmd" and returns the next one in the queue
        if any of this does not exist yet, wait for a RECV event until "wait" times out.
        RECV events are generated by the low-level recv thread "runEP5_recv()"
        '''
        startTime = time.time()
        self.recv_event.clear() # an event is only interesting if we've already failed to find our message

        while (time.time() - startTime)*1000 < wait:
            try:
                b = self.recv_mbox.get(app)
                if b:
                    if self._debug: print("Recv msg",app,b,cmd, file=sys.stderr)
                    if cmd is None:
                        keys = list(b.keys())
                        if len(keys):
                            cmd = list(b.keys())[-1] # just grab one.   no guarantees on the order

                if b is not None and cmd is not None:
                    q = b.get(cmd)
                    if self._debug: print("debug(recv) q='%s'"%repr(q), file=sys.stderr)

                    if q is not None and self.rsema.acquire(False):
                        if self._debug>3: print(("rsema.UNlocked", "rsema.locked")[self.rsema.locked()],2)
                        try:
                            resp, rt = q.pop(0)

                            self.rsema.release()
                            if self._debug>3: print(("rsema.UNlocked", "rsema.locked")[self.rsema.locked()],2)

                            # bring it on home...  this is the way out.
                            return resp[4:], rt

                        except IndexError:
                            pass

                        except AttributeError:
                            sys.excepthook(*sys.exc_info())
                            pass

                        self.rsema.release()

                self.recv_event.wait(old_div((wait - (time.time() - startTime)*1000),1000)) # wait on recv event, with timeout of remaining time
                self.recv_event.clear() # clear event, if it's set

            except KeyboardInterrupt:
                sys.excepthook(*sys.exc_info())
                break
            except:
                sys.excepthook(*sys.exc_info())

        raise ChipconUsbTimeoutException

    def recvAll(self, app, cmd=None):
        retval = self.recv_mbox.get(app,None)
        if retval is not None:
            if cmd is not None:
                b = retval
                if self.rsema.acquire():
                    #if self._debug: print ("rsema.UNlocked", "rsema.locked")[self.rsema.locked()],3
                    try:
                        retval = b.get(cmd)
                        b[cmd]=[]
                        if len(retval):
                            retval = [ (d[4:],t) for d,t in retval ] 
                    except:
                        sys.excepthook(*sys.exc_info())
                    finally:
                        self.rsema.release()
                        #if self._debug: print ("rsema.UNlocked", "rsema.locked")[self.rsema.locked()],3
            else:
                if self.rsema.acquire():
                    #if self._debug: print ("rsema.UNlocked", "rsema.locked")[self.rsema.locked()],4
                    try:
                        self.recv_mbox[app]={}
                    finally:
                        self.rsema.release()
                        #if self._debug: print ("rsema.UNlocked", "rsema.locked")[self.rsema.locked()],4
            return retval

    def send(self, app, cmd, buf, wait=USB_TX_WAIT):
        msg = b"%c%c%s%s" % (app, cmd, struct.pack("<H",len(buf)), buf)
        self.xsema.acquire()
        self.xmit_queue.append(msg)
        self.xmit_event.set()
        self.xsema.release()
        if self._debug: print("Sent Msg %s" %\
                hexlify(msg))
        return self.recv(app, cmd, wait)

    def reprDebugCodes(self, timeout=100):
        codes = self.getDebugCodes(timeout)
        if (codes != None and len(codes) == 2):
            rc1 = LCS.get(codes[0])
            rc2 = LCES.get(codes[0])
            return 'last position: %s\nlast error: %s' % (rc1, rc2)
        return codes

    def getDebugCodes(self, timeout=100):
        '''
        this function uses EP0 (not the normal USB EP5) to check the last state of the dongle.
        this only works if the dongle isn't in a hard-loop or some other corrupted state
        that neglects usbprocessing.

        two values are returned.  
        the first value is lastCode[0] and represents standard tracking messages (we were <here>)
        the second value is lastCode[1] and represents exception information (writing OUT while buffer in use!)

        messages LC_* and LCE_* (respectively) are defined in both global.h and rflib.chipcon_usb
        '''
        x = self._recvEP0(request=EP0_CMD_GET_DEBUG_CODES, timeout=timeout)
        if (x != None and len(x)==2):
            return struct.unpack("BB", x)
        else:
            return x

    def clearDebugCodes(self):
        retval = self.send(APP_SYSTEM, SYS_CMD_CLEAR_CODES, b"  ", 1000)
        return LCES.get(retval)

    def ep0GetAddr(self):
        addr = self._recvEP0(request=EP0_CMD_GET_ADDRESS)
        return addr
    def ep0Reset(self):
        x = self._recvEP0(request=0xfe, value=0x5352, index=0x4e54)
        return x

    def ep0Peek(self, addr, length, timeout=100):
        x = self._recvEP0(request=EP0_CMD_PEEKX, value=addr, length=length, timeout=timeout)
        return x#x[3:]

    def ep0Poke(self, addr, buf=b'\x00', timeout=100):
        x = self._sendEP0(request=EP0_CMD_POKEX, buf=buf, value=addr, timeout=timeout)
        return x

    def ep0Ping(self, count=10):
        good=0
        bad=0
        for x in range(count):
            #r = self._recvEP0(3, 10)
            try:
                r = self._recvEP0(request=2, value=count, length=count, timeout=DEFAULT_USB_TIMEOUT)
                print("PING: %d bytes received: %s"%(len(r), repr(r)))
            except ChipconUsbTimeoutException as e:
                r = None
                print("Ping Failed.",e)
            if r==None:
                bad+=1
            else:
                good+=1
        return (good,bad)

    def debug(self, delay=1):
        while True:
            """
            try:
                print(("DONGLE RESPONDING:  mode :%x, last error# %d"%(self.getDebugCodes()), file=sys.stderr)
            except:
                pass
            print('recv_queue:\t\t (%d bytes) "%s"'%(len(self.recv_queue),repr(self.recv_queue)[:len(self.recv_queue)%39+20]), file=sys.stderr)
            print('trash:     \t\t (%d bytes) "%s"'%(len(self.trash),repr(self.trash)[:len(self.trash)%39+20]), file=sys.stderr)
            print('recv_mbox  \t\t (%d keys)  "%s"'%(len(self.recv_mbox),repr(self.recv_mbox)[:len(repr(self.recv_mbox))%79]), file=sys.stderr)
            for x in self.recv_mbox.keys():
                print('    recv_mbox   %d\t (%d records)  "%s"'%(x,len(self.recv_mbox[x]),repr(self.recv_mbox[x])[:len(repr(self.recv_mbox[x]))%79]), file=sys.stderr)
                """
            print(self.reprRadioState())
            print(self.reprClientState())

            x,y,z = select.select([sys.stdin],[],[], delay)
            if sys.stdin in x:
                sys.stdin.read(1)
                break

    def getPartNum(self):
        try:
            r = self.send(APP_SYSTEM, SYS_CMD_PARTNUM, b"", 10000)
            r, rt = r
            return ord(r)

        except ChipconUsbTimeoutException as e:
            print("SETUP Failed.",e)
            return -1



    def ping(self, count=10, buf=b"ABCDEFGHIJKLMNOPQRSTUVWXYZ", wait=DEFAULT_USB_TIMEOUT, silent=False):
        good=0
        bad=0
        start = time.time()
        for x in range(count):
            istart = time.time()
            
            try:
                r = self.send(APP_SYSTEM, SYS_CMD_PING, buf, wait)
                r,rt = r
                istop = time.time()
                if not silent:
                    print("PING: %d bytes transmitted, received: %s (%f seconds)"%(len(buf), repr(r), istop-istart))
            except ChipconUsbTimeoutException as e:
                r = None
                if not silent:
                    print("Ping Failed.",e)
            if r==None:
                bad+=1
            else:
                good+=1
        stop = time.time()
        return (good,bad,stop-start)

    def bootloader(self):
        '''
        switch to bootloader mode. based on Fergus Noble's CC-Bootloader (https://github.com/fnoble/CC-Bootloader)
        this allows the firmware to be updated via USB instead of goodfet/ccdebugger
        '''
        try:
            self._bootloader = True
            r = self.send(APP_SYSTEM, SYS_CMD_BOOTLOADER, b"", wait=1)
        except ChipconUsbTimeoutException:
            pass
        
    def RESET(self):
        try:
            r = self.send(APP_SYSTEM, SYS_CMD_RESET, b"RESET_NOW\x00")
        except ChipconUsbTimeoutException:
            pass
        
    def peek(self, addr, bytecount=1):
        r, t = self.send(APP_SYSTEM, SYS_CMD_PEEK, struct.pack("<HH", bytecount, addr))
        return r

    def poke(self, addr, data):
        r, t = self.send(APP_SYSTEM, SYS_CMD_POKE, struct.pack("<H", addr) + data)
        return r
    
    def pokeReg(self, addr, data):
        r, t = self.send(APP_SYSTEM, SYS_CMD_POKE_REG, struct.pack("<H", addr) + data)
        return r

    def getBuildInfo(self):
        r, t = self.send(APP_SYSTEM, SYS_CMD_BUILDTYPE, b'')
        return r
            
    def getCompilerInfo(self):
        r, t = self.send(APP_SYSTEM, SYS_CMD_COMPILER, b'')
        return r

    def getDeviceSerialNumber(self):
        r, t = self.send(APP_SYSTEM, SYS_CMD_DEVICE_SERIAL_NUMBER, b'')
        return r
            
    def getInterruptRegisters(self):
        regs = {}
        # IEN0,1,2
        regs['IEN0'] = self.peek(IEN0,1)
        regs['IEN1'] = self.peek(IEN1,1)
        regs['IEN2'] = self.peek(IEN2,1)
        # TCON
        regs['TCON'] = self.peek(TCON,1)
        # S0CON
        regs['S0CON'] = self.peek(S0CON,1)
        # IRCON
        regs['IRCON'] = self.peek(IRCON,1)
        # IRCON2
        regs['IRCON2'] = self.peek(IRCON2,1)
        # S1CON
        regs['S1CON'] = self.peek(S1CON,1)
        # RFIF
        regs['RFIF'] = self.peek(RFIF,1)
        # DMAIE
        regs['DMAIE'] = self.peek(DMAIE,1)
        # DMAIF
        regs['DMAIF'] = self.peek(DMAIF,1)
        # DMAIRQ
        regs['DMAIRQ'] = self.peek(DMAIRQ,1)
        return regs

    def reprHardwareConfig(self):
        output= []

        hardware = self.getBuildInfo()
        output.append("Dongle:              %s" % hardware.split(b' ')[0])
        try:
            output.append("Firmware rev:        %s" % hardware.split(b'r')[1])
        except:
            output.append("Firmware rev:        Not found! Update needed!")
        try:
            compiler = self.getCompilerInfo()
            output.append("Compiler:            %s" % compiler)
        except:
            output.append("Compiler:            Not found! Update needed!")
        # see if we have a bootloader by loooking for it's recognition semaphores
        # in SFR I2SCLKF0 & I2SCLKF1
        if(self.peek(0xDF46,1) == b'\xF0' and self.peek(0xDF47,1) == b'\x0D'):
            output.append("Bootloader:          CC-Bootloader")
        else:
            output.append("Bootloader:          Not installed")
        return "\n".join(output)

    def reprSoftwareConfig(self):
        output= []

        output.append("rflib rev:           %s" % RFLIB_VERSION)
        return "\n".join(output)

    def printClientState(self, width=120):
        print(self.reprClientState(width))

    def reprClientState(self, width=120):
        output = ["="*width]
        output.append('     client thread cycles:      %d/%d' % (self.recv_threadcounter,self.send_threadcounter))
        output.append('     client errored cycles:     %d' % self._usberrorcnt)
        output.append('     recv_queue:                (%d bytes) %s'%(len(self.recv_queue),repr(self.recv_queue)[:width-42]))
        output.append('     trash:                     (%d blobs) "%s"'%(len(self.trash),repr(self.trash)[:width-44]))
        output.append('     recv_mbox                  (%d keys)  "%s"'%(len(self.recv_mbox),repr([hex(x) for x in list(self.recv_mbox.keys())])[:width-44]))
        for app in list(self.recv_mbox.keys()):
            appbox = self.recv_mbox[app]
            output.append('       app 0x%x (%d records)'%(app,len(appbox)))
            for cmd in list(appbox.keys()):
                output.append('             [0x%x]    (%d frames)  "%s"'%(cmd, len(appbox[cmd]), repr(appbox[cmd])[:width-36]))
            output.append('')
        return "\n".join(output)



def unittest(self, mhz=24):
    print("\nTesting USB ping()")
    self.ping(3)
    
    print("\nTesting USB ep0Ping()")
    self.ep0Ping()
    
    print("\nTesting USB enumeration")
    print("getString(1,100): %s" % repr(self._do.getString(1,100)))
    print("getString(2,100): %s" % repr(self._do.getString(2,100)))
    print("getString(3,100): %s" % repr(self._do.getString(3,100)))
    
    print("\nTesting USB EP MAX_PACKET_SIZE handling (ep0Peek(0xf000, 100))")
    print(repr(self.ep0Peek(0xf000, 100)))

    print("\nTesting USB EP MAX_PACKET_SIZE handling (peek(0xf000, 300))")
    print(repr(self.peek(0xf000, 400)))

    print("\nTesting USB poke/peek")
    data = b"".join([correctbytes(c) for c in range(120)])
    where = 0xf300
    self.poke(where, data)
    ndata = self.peek(where, len(data))
    if ndata != data:
        print(" *FAILED*\n '%s'\n '%s'" % (hexlify(data), hexlify(ndata)))
        raise Exception(" *FAILED*\n '%s'\n '%s'" % (hexlify(data), hexlify(ndata)))
    else:
        print("  passed  '%s'" % (hexlify(ndata)))


if __name__ == "__main__":
    idx = 0
    if len(sys.argv) > 1:
        idx = int(sys.argv.pop())
    d = USBDongle(idx=idx, debug=False)


